#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module containing windows, widgets etc. to create the QATS application

@author: perl
"""
import json
import logging
import os
import sys
from itertools import cycle

import numpy as np
from matplotlib.backends.backend_qt5agg import \
    FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import \
    NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
from pkg_resources import resource_filename
from qtpy import API_NAME as QTPY_API_NAME
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import (QAction, QCheckBox, QComboBox, QDialog,
                            QDialogButtonBox, QDoubleSpinBox, QFileDialog,
                            QFrame, QGroupBox, QHBoxLayout, QHeaderView,
                            QLabel, QLineEdit, QListView, QMainWindow,
                            QMessageBox, QPushButton, QRadioButton, QSpinBox,
                            QSplitter, QTabBar, QVBoxLayout, QWidget)

from ..stats.empirical import empirical_cdf
from ..tsdb import TsDB
from .funcs import (calculate_gumbel_fit, calculate_psd, calculate_rfc,
                    calculate_stats, calculate_trace, export_to_file,
                    import_from_file, read_timeseries)
from .logger import QLogger
from .threading import Worker
from .widgets import CustomTableWidget, CustomTableWidgetItem, CustomTabWidget
from .. import __version__

LOGGING_LEVELS = dict(
    debug=logging.DEBUG,
    info=logging.INFO,
    warning=logging.WARNING,
    error=logging.ERROR,
)
if sys.platform == "win32":
    SETTINGS_FILE = os.path.join(os.getenv("APPDATA", os.getenv("USERPROFILE", "")), "qats.settings")
else:
    SETTINGS_FILE = os.path.join("~", ".config", "qats.settings")
ICON_FILE = resource_filename("qats.app", "qats.ico")

STATS_ORDER = ["name", "min", "max", "mean", "std", "skew", "kurt", "tz", "wloc", "wscale", "wshape",
               "gloc", "gscale", "p_37.00", "p_57.00", "p_90.00"]
STATS_LABELS_TOOLTIPS = {
    "name": ("Name", "Time series name."),
    "min": ("Min.", "Sample minimum."),
    "max": ("Max.", "Sample maximum."),
    "mean": ("Mean", "Mean/average."),
    "std": ("Std.", "Unbiased standard deviation."),
    "skew": ("Skew.", "Skewness."),
    "kurt": ("Kurt.", "Kurtosis, Pearson’s definition (3.0 --> normal)."),
    "tz": ("Tz", "Average mean crossing period (s)."),
    "wloc": ("Wloc", "Weibull location parameter in distribution fitted to\n"
                     "sample maxima or -1 multiplied with the sample minima."),
    "wscale": ("Wscale", "Weibull scale parameter in distribution fitted to\n"
                         "sample maxima or -1 multiplied with the sample minima."),
    "wshape": ("Wshape", "Weibull shape parameter in distribution fitted to\n"
                         "sample maxima or -1 multiplied with the sample minima."),
    "gloc": ("Gloc", "Gumbel location parameter in sample extreme distribution,\n"
                     "derived from sample maxima/minima distribution."),
    "gscale": ("Gscale", "Gumbel location parameter in sample extreme distribution,\n"
                         "derived from sample maxima/minima distribution."),
    "p_37.00": ("P .37", "Most probable largest maximum (MPM). 37 percentile in\n"
                         "the extreme maxima/minima distribution. The generic\n"
                         "Gumbel (extreme value) distribution is derived from the Weibull\n"
                         "distribution fitted to sample maxima/minima."),
    "p_57.00": ("P .57", "Expected largest maximum. 57 percentile in\n"
                         "the extreme maxima/minima distribution. The generic\n"
                         "Gumbel (extreme value) distribution is derived from the Weibull\n"
                         "distribution fitted to sample maxima/minima."),
    "p_90.00": ("P .90", "90 percentile in the extreme maxima/minima distribution.\n"
                         "The generic Gumbel (extreme value) distribution is derived from the\n"
                         "Weibull distribution fitted to sample maxima/minima."),
}

# todo: New method that generalize threading
# todo: Explore how to create consecutive threads without handshake in main loop
# todo: add technical guidance and result interpretation to help menu, link docs website
# todo: add 'export' option to file menu: response statistics summary (mean, std, skew, kurt, tz, weibull distributions,
#  gumbel distributions etc.)


class Qats(QMainWindow):
    """
    Main window for the QATS application.

    Contain widgets for plotting time series, power spectra and statistics.

    Series of data are loaded from a time series file (e.g., .ts), and their names are displayed in a checkable 
    list view. The user can select the series it wants from the list and plot them on a matplotlib canvas. The 
    base library is used to load time series from file (`qats.io`), perform signal processing (`qats.signal`), 
    calculating power spectra and statistics (`qats.stats`) and plotting.
    """

    def __init__(self, parent=None, files_on_init=None, logging_level="info"):
        """
        Initiate main window

        Parameters
        ----------
        parent : QMainWindow, optional
            Parent window
        files_on_init : str|iterable, optional
            File names to be loaded on initiation
        logging_level : str, optional
            Logging level. Valid options: 'debug', 'info' (default), 'warning', 'error'.

        """
        super(Qats, self).__init__(parent)
        assert logging_level in LOGGING_LEVELS, "invalid logging level: '%s'" % logging_level

        # window title and icon (assumed located in 'images' at same level)
        self.setWindowTitle("QATS")
        self.icon = QIcon(ICON_FILE)
        self.setWindowIcon(self.icon)

        # create pool for managing threads
        self.threadpool = QThreadPool()

        # dictionary to hold settings
        self.settings_file = SETTINGS_FILE
        self.settings = dict()

        # load settings from previous sessions
        self.load_settings()

        # clipboard
        self.clip = QGuiApplication.clipboard()

        # create statusbar
        self.db_status = QLabel()
        self.statusBar().addPermanentWidget(self.db_status, stretch=0)

        # enable dropping url objects
        self.setAcceptDrops(True)

        # create frames
        self.main_frame = QFrame()
        self.setCentralWidget(self.main_frame)
        self.upper_left_frame = QFrame()
        self.bottom_left_frame = QFrame()
        self.left_frame = QFrame()
        self.right_frame = QFrame()

        # create tabs
        self.tabs = CustomTabWidget()

        # time domain plot tab
        w = QWidget()
        self.tabs.addTab(w, "Time history")
        self.tabs.setTabToolTip(0, "Plot data versus time for selected time series")
        self.tabs.tabBar().setTabButton(0, QTabBar.RightSide, None)     # disable close button
        self.history_fig = Figure()
        self.history_canvas = FigureCanvas(self.history_fig)
        self.history_canvas.setParent(w)
        self.history_axes = self.history_fig.add_subplot(111)
        self.history_mpl_toolbar = NavigationToolbar(self.history_canvas, self.upper_left_frame)
        vbox = QVBoxLayout()
        vbox.addWidget(self.history_canvas)
        vbox.addWidget(self.history_mpl_toolbar)
        w.setLayout(vbox)

        # power spectrum plot tab
        w = QWidget()
        self.tabs.addTab(w, "Power spectrum")
        self.tabs.setTabToolTip(1, "Plot power spectral density versus frequency for selected time series")
        self.tabs.tabBar().setTabButton(1, QTabBar.RightSide, None)  # disable close button
        self.spectrum_fig = Figure()
        self.spectrum_canvas = FigureCanvas(self.spectrum_fig)
        self.spectrum_canvas.setParent(w)
        self.spectrum_axes = self.spectrum_fig.add_subplot(111)
        self.spectrum_mpl_toolbar = NavigationToolbar(self.spectrum_canvas, self.upper_left_frame)
        vbox = QVBoxLayout()
        vbox.addWidget(self.spectrum_canvas)
        vbox.addWidget(self.spectrum_mpl_toolbar)
        w.setLayout(vbox)

        # statistics table
        w = QWidget()
        self.tabs.addTab(w, "Statistics")
        self.tabs.setTabToolTip(2, "Sample statistics for the selected time series")
        self.tabs.tabBar().setTabButton(2, QTabBar.RightSide, None)  # disable close button
        self.stats_table = CustomTableWidget()
        self.stats_table_initial_sort = None   # variable used to enable resetting of sorting order
        self.stats_table_initial_order = None  # variable used to enable resetting of sorting order
        vbox = QVBoxLayout()
        vbox.addWidget(self.stats_table)
        w.setLayout(vbox)
        self.reset_stats_table()

        # weibull paper plot tab
        w = QWidget()
        self.tabs.addTab(w, "Maxima/Minima CDF")
        self.tabs.setTabToolTip(3, "Plot fitted Weibull cumulative distribution function to maxima/minima of "
                                   "selected time series")
        self.tabs.tabBar().setTabButton(3, QTabBar.RightSide, None)  # disable close button
        self.weibull_fig = Figure()
        self.weibull_canvas = FigureCanvas(self.weibull_fig)
        self.weibull_canvas.setParent(w)
        self.weibull_axes = self.weibull_fig.add_subplot(111)
        self.weibull_mpl_toolbar = NavigationToolbar(self.weibull_canvas, self.upper_left_frame)
        vbox = QVBoxLayout()
        vbox.addWidget(self.weibull_canvas)
        vbox.addWidget(self.weibull_mpl_toolbar)
        w.setLayout(vbox)

        # cycle distribution plot tab
        w = QWidget()
        self.tabs.addTab(w, "Cycle distribution")
        self.tabs.setTabToolTip(4, "Plot distribution of cycle magnitude versus cycle count for "
                                   "selected time series")
        self.tabs.tabBar().setTabButton(4, QTabBar.RightSide, None)  # disable close button
        self.cycles_fig = Figure()
        self.cycles_canvas = FigureCanvas(self.cycles_fig)
        self.cycles_canvas.setParent(w)
        self.cycles_axes = self.cycles_fig.add_subplot(111)
        self.cycles_mpl_toolbar = NavigationToolbar(self.cycles_canvas, self.upper_left_frame)
        vbox = QVBoxLayout()
        vbox.addWidget(self.cycles_canvas)
        vbox.addWidget(self.cycles_mpl_toolbar)
        w.setLayout(vbox)

        # add layout to main frame (need to go via layout)
        main_hbox = QHBoxLayout()
        main_hbox.addWidget(self.tabs)
        self.upper_left_frame.setLayout(main_hbox)

        # initiate time series data base and checkable model and view with filter
        self.db = TsDB()
        self.db_source_model = QStandardItemModel()
        self.db_proxy_model = QSortFilterProxyModel()
        self.db_proxy_model.setDynamicSortFilter(True)
        self.db_proxy_model.setSourceModel(self.db_source_model)
        self.db_view = QListView()
        self.db_view.setModel(self.db_proxy_model)
        self.db_view_filter_casesensitivity = QCheckBox("Case sensitive filter")
        self.db_view_filter_casesensitivity.setChecked(False)
        self.db_view_filter_pattern = QLineEdit()
        self.db_view_filter_pattern.setPlaceholderText("type filter text")
        self.db_view_filter_pattern.setText("")
        self.db_view_filter_syntax = QComboBox()
        self.db_view_filter_syntax.addItem("Wildcard", "wildcard")
        self.db_view_filter_syntax.addItem("Regular expression", "regexp")
        self.db_view_filter_syntax.addItem("Fixed string", "fixedstring")
        self.db_view_filter_pattern.textChanged.connect(self.model_view_filter_changed)
        self.db_view_filter_syntax.currentIndexChanged.connect(self.model_view_filter_changed)
        self.db_view_filter_casesensitivity.toggled.connect(self.model_view_filter_changed)
        self.model_view_filter_changed()
        # unselect all items
        self.select_button = QPushButton("&Select all")
        self.unselect_button = QPushButton("&Unselect all")
        self.select_button.clicked.connect(self.on_select_all)
        self.unselect_button.clicked.connect(self.on_unselect_all)
        view_group = QGroupBox("Select time series")
        view_layout = QVBoxLayout()
        view_filter_hbox = QHBoxLayout()
        view_filter_hbox.addWidget(self.db_view_filter_pattern)
        view_filter_hbox.addWidget(self.db_view_filter_syntax)
        view_layout.addLayout(view_filter_hbox)
        view_layout.addWidget(self.db_view_filter_casesensitivity)
        view_select_hbox = QHBoxLayout()
        view_select_hbox.addWidget(self.select_button)
        view_select_hbox.addWidget(self.unselect_button)
        view_layout.addLayout(view_select_hbox)
        view_layout.addWidget(self.db_view)
        view_group.setLayout(view_layout)

        # time window selection
        time_group = QGroupBox("Set data processing time window")
        time_group.setToolTip("Calculations performed only for data within specified time window")
        ndecimals = self.twin_ndec()
        self.from_time = QDoubleSpinBox()  # time window
        self.to_time = QDoubleSpinBox()
        self.from_time.setRange(0, 1e12)
        self.to_time.setRange(0, 1e12)
        self.from_time.setEnabled(True)
        self.to_time.setEnabled(True)
        self.from_time.setSingleStep(10**(-ndecimals))
        self.to_time.setSingleStep(10**(-ndecimals))
        self.from_time.setSuffix(" s")
        self.to_time.setSuffix(" s")
        self.to_time.setDecimals(ndecimals)
        self.from_time.setDecimals(ndecimals)
        spins_hbox = QHBoxLayout()
        spins_hbox.addWidget(QLabel('from'))
        spins_hbox.addWidget(self.from_time)
        spins_hbox.addWidget(QLabel('to'))
        spins_hbox.addWidget(self.to_time)
        spins_hbox.addStretch(1)
        time_group.setLayout(spins_hbox)

        # set initial value of time window spin boxes
        self.from_time.setValue(0)
        self.to_time.setValue(1_000_000_000)  # 1_000_000

        # mutual exclusive peaks/troughs radio buttons
        minmax_group = QGroupBox("Select statistical quantity")
        minmax_group.setToolTip("Select maxima or minima as basis for the fitted and plotted cumulative"
                                " distribution functions. ")
        self.maxima = QRadioButton("Maxima")
        self.minima = QRadioButton("Minima")
        self.show_minmax = QCheckBox("Show in plot")
        self.maxima.setChecked(True)  # default maxima is checked
        minmax_hbox = QHBoxLayout()
        minmax_hbox.addWidget(self.maxima)
        minmax_hbox.addWidget(self.minima)
        minmax_hbox.addWidget(self.show_minmax)
        minmax_hbox.addStretch(1)
        minmax_group.setLayout(minmax_hbox)

        # filter selection and frequency window
        self.no_filter = QRadioButton("None")
        self.no_filter.setCheckable(False)
        self.no_filter.toggled.connect(self.on_no_filter)
        no_filter_hbox = QHBoxLayout()
        no_filter_hbox.addWidget(self.no_filter)

        self.lowpass = QRadioButton("Low-pass")
        self.lowpass.setCheckable(False)
        self.lowpass.toggled.connect(self.on_lowpass)
        self.lowpass_f = QDoubleSpinBox()
        self.lowpass_f.setEnabled(False)  # default opaque
        lowpass_hbox = QHBoxLayout()
        lowpass_hbox.addWidget(self.lowpass)
        lowpass_hbox.addStretch(1)
        lowpass_hbox.addWidget(QLabel("below"))
        lowpass_hbox.addWidget(self.lowpass_f)

        self.hipass = QRadioButton("High-pass")
        self.hipass.setCheckable(False)
        self.hipass.toggled.connect(self.on_hipass)
        self.hipass_f = QDoubleSpinBox()
        self.hipass_f.setEnabled(False)  # default opaque
        hipass_hbox = QHBoxLayout()
        hipass_hbox.addWidget(self.hipass)
        hipass_hbox.addStretch(1)
        hipass_hbox.addWidget(QLabel("above"))
        hipass_hbox.addWidget(self.hipass_f)

        self.bandpass = QRadioButton("Band-pass")
        self.bandpass.setCheckable(False)
        self.bandpass.toggled.connect(self.on_bandpass)
        self.bandpass_lf = QDoubleSpinBox()
        self.bandpass_hf = QDoubleSpinBox()
        self.bandpass_lf.setEnabled(False)  # default opaque
        self.bandpass_hf.setEnabled(False)  # default opaque
        bandpass_hbox = QHBoxLayout()
        bandpass_hbox.addWidget(self.bandpass)
        bandpass_hbox.addStretch(1)
        bandpass_hbox.addWidget(QLabel("between"))
        bandpass_hbox.addWidget(self.bandpass_lf)
        bandpass_hbox.addWidget(QLabel("and"))
        bandpass_hbox.addWidget(self.bandpass_hf)

        self.bandblock = QRadioButton("Band-block")
        self.bandblock.setCheckable(False)
        self.bandblock.toggled.connect(self.on_bandblock)
        self.bandblock_lf = QDoubleSpinBox()
        self.bandblock_hf = QDoubleSpinBox()
        self.bandblock_lf.setEnabled(False)  # default opaque
        self.bandblock_hf.setEnabled(False)  # default opaque
        bandblock_hbox = QHBoxLayout()
        bandblock_hbox.addWidget(self.bandblock)
        bandblock_hbox.addStretch(1)
        bandblock_hbox.addWidget(QLabel("between"))
        bandblock_hbox.addWidget(self.bandblock_lf)
        bandblock_hbox.addWidget(QLabel("and"))
        bandblock_hbox.addWidget(self.bandblock_hf)

        # set range, decimals and suffix of frequency filter range spin boxes
        for w in [self.lowpass_f, self.hipass_f, self.bandpass_lf, self.bandpass_hf, self.bandblock_lf,
                  self.bandblock_hf]:
            w.setRange(0.0, 50.)
            w.setDecimals(3)
            w.setSuffix(" Hz")

        # make filter selection radio buttons checkable
        for w in [self.no_filter, self.lowpass, self.hipass, self.bandpass, self.bandblock]:
            w.setCheckable(True)

        # un-filtered data series by default
        self.no_filter.toggle()

        # stack radio-butttons and spin-boxes vertically
        filter_vbox = QVBoxLayout()
        for hb in [no_filter_hbox, lowpass_hbox, hipass_hbox, bandpass_hbox, bandblock_hbox]:
            filter_vbox.addLayout(hb)

        filter_group = QGroupBox("Apply frequency filter")
        filter_group.setLayout(filter_vbox)

        # show plots / update GUI
        self.display_button = QPushButton("&Display")
        self.display_button.clicked.connect(self.on_display)

        # create right hand vertical box layout and add data series check list,
        # time window spin boxes, show button, legend check box
        right_vbox = QVBoxLayout()
        right_vbox.addWidget(view_group)
        right_vbox.addWidget(time_group)
        right_vbox.addWidget(minmax_group)
        right_vbox.addWidget(filter_group)
        right_vbox.addWidget(self.display_button)

        # add layout to right hand frame
        self.right_frame.setLayout(right_vbox)

        # create logger widget
        self.logger = QLogger(self)
        self.logger.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
        logging.getLogger().addHandler(self.logger)
        logging.getLogger().setLevel(LOGGING_LEVELS[logging_level])
        logger_layout = QVBoxLayout()
        logger_layout.addWidget(self.logger.widget)
        self.bottom_left_frame.setLayout(logger_layout)

        # add adjustable splitter to main window
        vertical_splitter = QSplitter(Qt.Vertical)
        vertical_splitter.addWidget(self.upper_left_frame)
        vertical_splitter.addWidget(self.bottom_left_frame)
        vbox = QVBoxLayout()
        vbox.addWidget(vertical_splitter)
        self.left_frame.setLayout(vbox)
        horizontal_splitter = QSplitter(Qt.Horizontal)
        horizontal_splitter.addWidget(self.left_frame)
        horizontal_splitter.addWidget(self.right_frame)
        hbox = QHBoxLayout()
        hbox.addWidget(horizontal_splitter)
        self.main_frame.setLayout(hbox)

        # create menubar, file menu and help menu
        self.menu_bar = self.menuBar()
        file_menu = self.menu_bar.addMenu("&File")
        tool_menu = self.menu_bar.addMenu("&Tools")
        help_menu = self.menu_bar.addMenu("&Help")

        # create File menu and actions
        import_action = QAction("Import from file", self)
        import_action.setShortcut("Ctrl+I")
        import_action.setStatusTip("Import time series from file")
        import_action.triggered.connect(self.on_import)

        export_action = QAction("Export to file", self)
        export_action.setShortcut("Ctrl+E")
        export_action.setStatusTip("Export time series to file")
        export_action.triggered.connect(self.on_export)

        clear_action = QAction("Clear", self)
        clear_action.setShortcut("Ctrl+Del")
        clear_action.setStatusTip("Clear all time series from database")
        clear_action.triggered.connect(self.on_clear)

        clear_log_action = QAction("Clear logger", self)
        clear_log_action.setShortcut("Ctrl+Shift+Del")
        clear_log_action.setStatusTip("Clear logger widget")
        clear_log_action.triggered.connect(self.logger.clear)

        settings_action = QAction("Settings", self)
        settings_action.setShortcut("Ctrl+Shift+S")
        settings_action.setStatusTip("Configure application settings")
        settings_action.triggered.connect(self.on_open_settings)

        quit_action = QAction("&Quit", self)
        quit_action.setShortcut("Ctrl+Q")
        quit_action.setToolTip("Close the application")
        quit_action.triggered.connect(self.close)

        plot_gumbel_action = QAction("Plot extremes CDF", self)
        plot_gumbel_action.setToolTip("Plot fitted Gumbel cumulative distribution function to extremes"
                                      " in selected time series")
        plot_gumbel_action.triggered.connect(self.on_create_gumbel_plot)

        about_action = QAction("&About", self)
        about_action.setShortcut("F1")
        about_action.setToolTip("About the application")
        about_action.triggered.connect(self.on_about)

        file_menu.addAction(import_action)
        file_menu.addAction(export_action)
        file_menu.addAction(clear_action)
        file_menu.addAction(clear_log_action)
        file_menu.addAction(settings_action)
        file_menu.addSeparator()
        file_menu.addAction(quit_action)
        tool_menu.addAction(plot_gumbel_action)
        help_menu.addAction(about_action)

        # load files specified on initation
        if files_on_init is not None:
            if isinstance(files_on_init, str):
                self.load_files([files_on_init])
            elif isinstance(files_on_init, tuple) or isinstance(files_on_init, list):
                self.load_files(files_on_init)

        # refresh
        self.reset_axes()
        self.set_status(message="Welcome! Please load a file to get started.")

    @staticmethod
    def log_thread_exception(exc):
        """
        Pipe exceptions from threads other than main thread to logger

        Parameters
        ----------
        exc : tuple
            Exception type, value and traceback

        """
        # choose to pipe only the the exception value, not type nor full traceback
        logging.error("%s - %s" % exc[1:])

    def closeEvent(self, event):
        """Overload close event to save settings before exit."""
        self.save_settings()

    def dragEnterEvent(self, event):
        """
        Event handler for dragging objects over main window. Overrides method in QWidget.
        """
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """
        Event handler for dropping objects over main window. Overrides method in QWidget.
        """
        files = [u.toLocalFile() for u in event.mimeData().urls()]
        self.load_files(files)

    def filter_settings(self):
        """
        Return filter type and cut off frequencies
        """
        if self.lowpass.isChecked():
            args = ('lp', self.lowpass_f.value())
        elif self.hipass.isChecked():
            args = ('hp', self.hipass_f.value())
        elif self.bandpass.isChecked():
            args = ('bp', self.bandpass_lf.value(), self.bandpass_hf.value())
        elif self.bandblock.isChecked():
            args = ('bs', self.bandblock_lf.value(), self.bandblock_hf.value())
        else:
            args = None

        return args

    def keyPressEvent(self, e):
        selected = self.stats_table.selectedRanges()
        if e.key() == Qt.Key_C:  # Ctr+C
            s = "\t".join([str(self.stats_table.horizontalHeaderItem(i).text()) for i in
                                  range(selected[0].leftColumn(), selected[0].rightColumn() + 1)]) + "\n"

            for r in range(selected[0].topRow(), selected[0].bottomRow() + 1):
                for c in range(selected[0].leftColumn(), selected[0].rightColumn() + 1):
                    try:
                        s += str(self.stats_table.item(r, c).text()) + "\t"
                    except AttributeError:
                        s += "\t"
                s = s[:-1] + "\n"  # eliminate last '\t'
            self.clip.setText(s)

    def load_files(self, files):
        """
        Load files into application

        Parameters
        ----------
        files : list
            list of file names

        """

        if len(files):
            # update statusbar
            self.set_status("Importing %d file(s)...." % len(files))

            # Pass the function to execute, args, kwargs are passed to the run function
            worker = Worker(import_from_file, list(files))

            # pipe exceptions to logger (NB: like this because logging module cannot be used in pyqt QThreads)
            worker.signals.error.connect(self.log_thread_exception)

            # grab results and merge into model
            worker.signals.result.connect(self.update_model)

            # update model and status bar once finished
            worker.signals.finished.connect(self.set_status)

            # Execute
            self.threadpool.start(worker)

    def load_settings(self):
        """Load settings from file."""
        try:
            with open(self.settings_file) as fp:
                self.settings = json.load(fp)
        except FileNotFoundError:
            self.settings = dict()

    def model_view_filter_changed(self):
        """
        Apply filter changes to db proxy model
        """
        # type of filter selected in drop-down menu
        filter_index = self.db_view_filter_syntax.currentIndex()
        filter_type = self.db_view_filter_syntax.itemData(filter_index)
        # text input by user
        pattern = self.db_view_filter_pattern.text()
        # case sensitivity (if 'Case sensitive filter' is checked)
        case_sensitive = self.db_view_filter_casesensitivity.isChecked()
        
        # the code below works for python qt5 (pyside2/pyqt5) and qt6 (pyside6/pyqt6)
        # pyside6: see the following links for documentation on QRegularExpression and the filter model (QSortFilterProxyModel)
        #   https://doc.qt.io/qtforpython-6/PySide6/QtCore/QRegularExpression.html
        #   https://doc-snapshots.qt.io/qtforpython-6.2/PySide6/QtCore/QSortFilterProxyModel.html#filtering
        #   https://doc-snapshots.qt.io/qtforpython-6.2/PySide6/QtCore/QSortFilterProxyModel.html#PySide6.QtCore.QSortFilterProxyModel.filterAcceptsRow

        # notes on the methods available for self.db_proxy_model (type: QSortFilterProxyModel)
        #   .setFilterCaseSensitivity(Qt.CaseSensitive | Qt.CaseInsensitive) may be used with .setFilterWildcard(pattern) and .setFilterFixedString(pattern)
        #   .setFilterRegularExpression(QRegularExpression) may not be used with .setFilterCaseSensitivity(...)
        #       * setting a new regular expression propagates its case sensitivity to .filterCaseSensitivity (-> breaks the binding to what previously set)
        #       * setting a filter case sensitivity afterwards breaks the binding to the regular expression
        
        # construct regexp string that may be used to initiate QRegularExpression instance
        if filter_type == "wildcard":
            # pad with wildcard ('*') to get expected behaviour
            reg_exp_pattern = QRegularExpression.wildcardToRegularExpression(f"*{pattern}*")
        elif filter_type == "regexp":
            # pattern string should be interpreted as a regexp pattern
            reg_exp_pattern = pattern
        elif filter_type == "fixedstring":
            # according to https://doc.qt.io/qt-6/qregexp.html, a fixed string is 
            # equivalent to using the regexp pattern on a string in which all 
            # metacharacters are escaped using escape()
            reg_exp_pattern = QRegularExpression.escape(pattern)
        else:
            raise ValueError(f"Unsupported filter type: {filter_type}")
    
        # options (for case sensitivity) to QRegularExpression construction
        if case_sensitive:
            # case sensitive is default for QRegularExpression => no options
            options = {}
        else:
            # case insensitive => must be specified
            options = {"options": QRegularExpression.CaseInsensitiveOption}

        # initiate QRegularExpression instance
        reg_exp = QRegularExpression(reg_exp_pattern, **options)
        
        # assign reg exp to proxy model filter
        # (but only if reg exp is valid, which is not always the case when user is still typing)
        if reg_exp.isValid():
            self.db_proxy_model.setFilterRegularExpression(reg_exp)

    def on_about(self):
        """
        Show information about the application
        """
        msg = "This is a low threshold tool for inspection of time series, power spectra and statistics. " \
              "Its main objective is to ease self-check, quality assurance and reporting.<br><br>" \
              "Import qats Python package and use the <a href='https://qats.readthedocs.io/en/latest/'>API</a> " \
              "when you need advanced features or want to extend it's functionality.<br><br>" \
              "Please send feature requests, technical queries and bug reports to the developers on " \
              "<a href='https://github.com/dnvgl/qats/issues'>Github</a>.<br><br>" \
              "ENJOY! <br><br>" \
              f"QT API used: {QTPY_API_NAME}"

        msgbox = QMessageBox()
        msgbox.setWindowIcon(self.icon)
        msgbox.setIcon(QMessageBox.Information)
        msgbox.setTextFormat(Qt.RichText)
        msgbox.setText(msg.strip())
        msgbox.setWindowTitle(f"About QATS - version {__version__}")
        msgbox.exec_()

    def on_clear(self):
        """
        Clear all time series from database
        """
        self.db.clear(names="*", display=False)
        self.db_source_model.clear()
        self.reset_axes()
        self.reset_stats_table()
        logging.info("Cleared all time series from database...")
        self.set_status()

    def on_create_gumbel_plot(self):
        """
        Create new tab with canvas and plot extremes sample on Gumbel scales
        """

        # list of selected series
        selected_series = self.selected_series()

        if len(selected_series) > 1:
            # update statusbar
            self.set_status("Reading time series...", msecs=10000)  # will probably be erased by new status message

            # Pass the function to execute, args, kwargs are passed to the run function
            # todo: consider if it is necessary to pass copied db to avoid main loop freeze
            worker = Worker(read_timeseries, self.db, selected_series)

            # pipe exceptions to logger (NB: like this because logging module cannot be used in pyqt QThreads)
            worker.signals.error.connect(self.log_thread_exception)

            # grab results start further calculations
            worker.signals.result.connect(self.start_gumbel_fit_thread)

            # Execute
            self.threadpool.start(worker)
        else:
            # inform user to select at least one time series before plotting
            logging.info("Select more than 1 time series to fit a Gumbel distribution to sample of extremes.")

    def on_display(self):
        """
        Plot checked data series when pressing the 'show' button.
        """
        self.reset_stats_table()  # TODO: Rework the gui logic and implement an overall reseti()
        # list of selected series
        selected_series = self.selected_series()

        if len(selected_series) >= 1:
            # update statusbar
            self.set_status("Reading time series...", msecs=10000)  # will probably be erased by new status message

            # Pass the function to execute, args, kwargs are passed to the run function
            # todo: consider if it is necessary to pass copied db to avoid main loop freeze
            worker = Worker(read_timeseries, self.db, selected_series)

            # pipe exceptions to logger (NB: like this because logging module cannot be used in pyqt QThreads)
            worker.signals.error.connect(self.log_thread_exception)

            # grab results start further calculations
            worker.signals.result.connect(self.start_times_series_processing_thread)

            # Execute
            self.threadpool.start(worker)

        else:
            # inform user to select at least one time series before plotting
            logging.info("Select at least 1 time series before plotting.")

    def on_export(self):
        """
        Export selected time series to file
        """
        # file save dialogue
        dlg = QFileDialog()
        dlg.setWindowIcon(self.icon)
        options = dlg.Options()

        name, _ = dlg.getSaveFileName(dlg, "Export time series to file", "",
                                      "Direct access file (*.ts);;"
                                      "ASCII file with header (*.dat);;"
                                      "SIMA H5 file (*.h5);;"
                                      "All Files (*)", options=options)

        # get list of selected time series
        keys = self.selected_series()

        # get ui settings
        fargs = self.filter_settings()
        twin = self.time_window()

        if name:    # nullstring if file dialog is cancelled
            # update statusbar
            self.set_status("Exporting....")

            # Pass the function to execute, args, kwargs are passed to the run function
            worker = Worker(export_to_file, name, self.db, keys, twin, fargs)

            # pipe exceptions to logger
            worker.signals.error.connect(self.log_thread_exception)

            # update status bar once finished
            worker.signals.finished.connect(self.set_status)

            # Execute
            self.threadpool.start(worker)

    def on_import(self):
        """
        File open dialogue
        """
        dlg = QFileDialog()
        dlg.setWindowIcon(self.icon)
        options = dlg.Options()
        files, _ = dlg.getOpenFileNames(dlg, "Load time series files", "",
                                        "Direct access files (*.ts);;"
                                        "SIMO S2X direct access files with info array (*.tda);;"
                                        "RIFLEX SIMO binary files (*.bin);;"
                                        "RIFLEX SIMO ASCII files (*.asc);;"
                                        "SINTEF Ocean test data export format (*.mat);;"
                                        "ASCII file with header (*.dat);;"
                                        "SIMA H5 files (*.h5);;"
                                        "CSV file with header (*.csv);;"
                                        "Technical Data Management Streaming files (*.tdms);;"
                                        "All Files (*)", options=options)

        # load files into db and update application model and view
        self.load_files(files)

    def on_no_filter(self):
        """
        Toggle off all filters and disable spin boxes
        """
        if self.db.n > 0:
            self.lowpass_f.setEnabled(False)
            self.hipass_f.setEnabled(False)
            self.bandpass_lf.setEnabled(False)
            self.bandpass_hf.setEnabled(False)
            self.bandblock_lf.setEnabled(False)
            self.bandblock_hf.setEnabled(False)

    def on_lowpass(self):
        """
        Toggle off filters and disable spin boxes, except low-pass
        """
        if self.db.n > 0:
            self.lowpass_f.setEnabled(True)
            self.hipass_f.setEnabled(False)
            self.bandpass_lf.setEnabled(False)
            self.bandpass_hf.setEnabled(False)
            self.bandblock_lf.setEnabled(False)
            self.bandblock_hf.setEnabled(False)

    def on_hipass(self):
        """
        Toggle off filters and disable spin boxes, except high-pass
        """
        if self.db.n > 0:
            self.lowpass_f.setEnabled(False)
            self.hipass_f.setEnabled(True)
            self.bandpass_lf.setEnabled(False)
            self.bandpass_hf.setEnabled(False)
            self.bandblock_lf.setEnabled(False)
            self.bandblock_hf.setEnabled(False)

    def on_bandpass(self):
        """
        Toggle off filters and disable spin boxes, except band-pass
        """
        if self.db.n > 0:
            self.lowpass_f.setEnabled(False)
            self.hipass_f.setEnabled(False)
            self.bandpass_lf.setEnabled(True)
            self.bandpass_hf.setEnabled(True)
            self.bandblock_lf.setEnabled(False)
            self.bandblock_hf.setEnabled(False)

    def on_bandblock(self):
        """
        Toggle off filters and disable spin boxes, except band-block
        """
        if self.db.n > 0:
            self.lowpass_f.setEnabled(False)
            self.hipass_f.setEnabled(False)
            self.bandpass_lf.setEnabled(False)
            self.bandpass_hf.setEnabled(False)
            self.bandblock_lf.setEnabled(True)
            self.bandblock_hf.setEnabled(True)

    def on_select_all(self):
        """
        Check all items in item model
        """
        for row_number in range(self.db_proxy_model.rowCount()):
            proxy_index = self.db_proxy_model.index(row_number, 0)
            source_index = self.db_proxy_model.mapToSource(proxy_index)
            item = self.db_source_model.itemFromIndex(source_index)
            item.setCheckState(Qt.Checked)

    def on_open_settings(self):
        """
        Configure the application settings.
        """
        # load settings dialog with current settings
        defaults = (self.psd_normalized(), self.psd_nperseg(), self.rfc_nbins(), self.twin_ndec())
        psdnorm, nperseg, nbins, twindec, ok = SettingsDialog.settings(defaults, parent=self)

        # update settings
        if ok:
            self.settings["psd_normalized"] = psdnorm
            self.settings["psd_nperseg"] = nperseg
            self.settings["rfc_nbins"] = nbins
            self.settings["twin_ndec"] = twindec

    def on_unselect_all(self):
        """
        Uncheck all items in item model
        """
        for row_number in range(self.db_proxy_model.rowCount()):
            proxy_index = self.db_proxy_model.index(row_number, 0)
            source_index = self.db_proxy_model.mapToSource(proxy_index)
            item = self.db_source_model.itemFromIndex(source_index)
            item.setCheckState(Qt.Unchecked)

    def plot_trace(self, container):
        """
        Plot time series trace and peaks/troughs

        Parameters
        ----------
        container : dict
            Time, data, peak and trough values
        """
        # clear axes
        self.history_axes.clear()
        self.history_axes.grid(True)
        self.history_axes.set_xlabel('Time (s)')

        # draw
        for name, data in container.items():
            # plot timetrace
            self.history_axes.plot(data.get('t'), data.get('x'), '-', label=name)

            # include maxima/minima if requested
            if self.show_minmax.isChecked() and self.maxima.isChecked():
                # maxima
                self.history_axes.plot(data.get('tmax'), data.get('xmax'), 'o')
            elif self.show_minmax.isChecked() and self.minima.isChecked():
                # minima
                self.history_axes.plot(data.get('tmin'), data.get('xmin'), 'o')

            self.history_axes.legend(loc="upper left")
            self.history_canvas.draw()

        self.set_status("History plot updated", msecs=3000)

    def plot_psd(self, container):
        """
        Plot time series power spectral density

        Parameters
        ----------
        container : dict
            Frequency versus power spectral density as tuple
        """
        # clear axes
        self.spectrum_axes.clear()
        self.spectrum_axes.grid(True)
        self.spectrum_axes.set_xlabel('Frequency (Hz)')
        self.spectrum_axes.set_ylabel('Power spectral density')

        # draw
        for name, value in container.items():
            f, s = value
            self.spectrum_axes.plot(f, s, '-', label=name)
            self.spectrum_axes.legend(loc="upper left")
            self.spectrum_canvas.draw()

        self.set_status("Power spectral density plot updated", msecs=3000)

    def plot_rfc(self, container):
        """
        Plot cycle ranges versus number of occurrences

        Parameters
        ----------
        container : dict
            Cycle range versus number of occurrences.
        """
        self.cycles_axes.clear()
        self.cycles_axes.grid(True)
        self.cycles_axes.set_xlabel('Cycle range')
        self.cycles_axes.set_ylabel('Cycle count (-)')

        # cycle bar colors
        barcolor = cycle("bgrcmyk")

        # draw
        for name, value in container.items():
            crange, count = value    # unpack magnitude and count

            try:
                # width of bars
                width = crange[1] - crange[0]

            except IndexError:
                # cycles and magnitude lists are empty, no cycles found from rainflow
                logging.warning("No cycles found for time series '%s'. Cannot create cycle histogram." % name)

            except ValueError:
                # probably nans or infs in data
                logging.warning("Invalid values (nan, inf) in time series '%s'. Cannot create cycle histogram." % name)

            else:
                self.cycles_axes.bar(crange, count, width, label=name, alpha=0.4, color=next(barcolor))
                self.cycles_axes.legend(loc="upper left")
                self.cycles_canvas.draw()

        self.set_status("Cycle distribution plot updated", msecs=3000)

    def plot_weibull(self, container):
        """
        Plot maxima/minima sample on linearized weibull axes

        Parameters
        ----------
        container : dict
            Sample and fitted weibull parameters
        """
        self.weibull_axes.clear()
        self.weibull_axes.grid(True)
        self.weibull_axes.set_xlabel('X - location')
        self.weibull_axes.set_ylabel('Cumulative probability (-)')

        # labels and tick positions for weibull paper plot
        p_labels = np.array([0.2, 0.5, 0.7, 0.8, 0.9, 0.95, 0.99, 0.999, 0.9999])
        p_ticks = np.log(np.log(1. / (1. - p_labels)))
        x_lb, x_ub = None, None

        # draw
        for name, value in container.items():
            x = value.get("sample")
            loc = value.get("wloc")
            scale = value.get("wscale")
            shape = value.get("wshape")
            is_minima = value.get("is_minima")

            if x is None:
                # skip
                continue

            # flip sample to be able to plot sample on weibull scales
            if is_minima:
                x *= -1.

            # normalize maxima/minima sample on weibull scales
            x = np.sort(x)  # sort ascending
            mask = (x >= loc)  # weibull paper plot will fail for mv-loc < 0
            x_norm = np.log(x[mask] - loc)
            ecdf_norm = np.log(np.log(1. / (1. - (np.arange(x.size) + 1.) / (x.size + 1.))))
            q_fitted = scale * (-np.log(1. - p_labels)) ** (1. / shape)  # x-loc

            # consider switching to np.any(), not sure what is more correct
            if np.all(q_fitted <= 0.):
                logging.warning("Invalid sample for time series '%s'. Cannot fit Weibull distribution." % name)

            else:
                # normalized quantiles from fitted distribution (inside if-statement to avoid log(negative_num)
                q_norm_fitted = np.log(q_fitted)

                # calculate data range for xtick/label calculation later
                if not x_lb:
                    # first time
                    x_lb = np.min(q_fitted)
                elif np.min(q_fitted) < x_lb:
                    # lower value
                    x_lb = np.min(q_fitted)

                if not x_ub:
                    x_ub = np.max(q_fitted)
                elif np.max(q_fitted) > x_ub:
                    x_ub = np.max(q_fitted)

                # calculate axes tick and labels
                labels_sample = np.around(np.linspace(x_lb, x_ub, 4), decimals=1)

                ticks_sample = np.log(labels_sample[labels_sample > 0.])

                # and draw weibull paper plot (avoid log(0))
                self.weibull_axes.plot(x_norm, ecdf_norm[mask], 'o', label=name)
                self.weibull_axes.plot(q_norm_fitted, p_ticks, '-')

                self.weibull_axes.set_xticks(ticks_sample)
                if self.maxima.isChecked():
                    self.weibull_axes.set_xticklabels(labels_sample[labels_sample > 0.])
                else:
                    self.weibull_axes.set_xticklabels(-1. * labels_sample[labels_sample > 0.])

                self.weibull_axes.set_ylim((p_labels[0], p_labels[-1]))
                self.weibull_axes.set_yticks(p_ticks)
                self.weibull_axes.set_yticklabels(p_labels)
                self.weibull_axes.legend(loc='upper left')
                self.weibull_canvas.draw()

            self.set_status("Weibull distribution plot updated", msecs=3000)

    def plot_gumbel(self, container):
        """
        Create new closable tab widget with plot of fitted Gumbel cumulative distribution function to extremes
        of selected time series

        Parameters
        ----------
        container : dict
            Sample and fitted Gumbel distribution parameters
        """
        # get sample and fitted distribution parameters
        sample = container.get('sample')
        loc = container.get('loc')
        scale = container.get('scale')
        is_minima = container.get('minima')

        # nomalize sample distribution and fitted distribution
        sample_dist = -np.log(-np.log(empirical_cdf(sample.size, kind="median")))
        fitted_dist = (sample - loc) / scale

        # create widget and attach to tab
        w = QWidget()
        fig = Figure()
        canvas = FigureCanvas(fig)
        canvas.setParent(w)
        axes = fig.add_subplot(111)
        toolbar = NavigationToolbar(canvas, self.upper_left_frame)
        vbox = QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        w.setLayout(vbox)
        self.tabs.addTab(w, "Extremes CDF")
        tabindex = self.tabs.indexOf(w)
        self.tabs.setTabToolTip(tabindex, "Plot fitted Gumbel cumulative distribution function "
                                          "to extremes (maxima/minima) of selected time series")
        if is_minima:
            # TODO: Double check that this is correct considering sample multiplied with -1 etc.
            axes.invert_xaxis()
            txt = 'minima'
        else:
            txt = 'maxima'

        # plot
        # TODO: Double check if sample should be multiplied with -1 or not.
        axes.plot(sample, sample_dist, 'ko', label='Data')
        axes.plot(sample, fitted_dist, '-m', label='Fitted')

        # plotting positions and plot configurations
        ylabels = np.array([0.1, 0.2, 0.5, 0.7, 0.8, 0.9, 0.95, 0.99, 0.999])
        yticks = -np.log(-np.log(ylabels))
        axes.set_yticks(yticks)
        axes.set_yticklabels(ylabels)
        axes.legend(loc="upper left")
        axes.grid(True)
        axes.set_xlabel("Data")
        axes.set_ylabel("Cumulative probability (-)")
        canvas.draw()

        self.set_status("Gumbel distribution plot created", msecs=3000)
        logging.info(f"Fitted Gumbel distribution to {txt} extreme sample of {sample.size}'. "
                     f"(location, scale) = ({loc}, {scale})")

    def psd_nperseg(self):
        """int: Length of segments used to estimate PSD with Welch's method."""
        return self.settings.get("psd_nperseg", 20000)

    def psd_normalized(self):
        """bool: Plot PSD normalized by sample variance or not."""
        return self.settings.get("psd_normalized", False)

    def rfc_nbins(self):
        """int: Number of bins in cycle distribution."""
        return self.settings.get("rfc_nbins", 256)

    def twin_ndec(self):
        """int: Number of decimals in time window for data processing."""
        return self.settings.get("twin_ndec", 2)

    def reset_axes(self):
        """
        Clear and reset plot axes
        """
        self.history_axes.clear()
        self.history_axes.grid(True)
        self.history_axes.set_xlabel('Time (s)')
        self.history_canvas.draw()
        self.spectrum_axes.clear()
        self.spectrum_axes.grid(True)
        self.spectrum_axes.set_xlabel('Frequency (Hz)')
        self.spectrum_axes.set_ylabel('Spectral density')
        self.spectrum_canvas.draw()
        self.weibull_axes.clear()
        self.weibull_axes.grid(True)
        self.weibull_axes.set_xlabel('X - location')
        self.weibull_axes.set_ylabel('Cumulative probability (-)')
        self.weibull_canvas.draw()
        self.cycles_axes.clear()
        self.cycles_axes.grid(True)
        self.cycles_axes.set_xlabel('Cycle magnitude')
        self.cycles_axes.set_ylabel('Cycle count (-)')
        self.cycles_canvas.draw()

    def reset_stats_table(self):
        """Reset statistics table."""
        # re-create empty table with header
        self.stats_table.setRowCount(0)
        self.stats_table.setColumnCount(len(STATS_ORDER))
        self.stats_table.setAlternatingRowColors(True)
        header_labels = [STATS_LABELS_TOOLTIPS.get(k, [k, None])[0] for k in STATS_ORDER]
        self.stats_table.setHorizontalHeaderLabels(header_labels)
        for i, k in enumerate(STATS_ORDER):  # set tooltips for column headers
            tooltip = STATS_LABELS_TOOLTIPS.get(k, [None, None])[0]
            if tooltip is not None:
                self.stats_table.horizontalHeaderItem(i).setToolTip(tooltip)
        header = self.stats_table.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.Interactive)
        self.stats_table.verticalHeader().setVisible(False)

    def save_settings(self):
        """Save settings to file."""
        # create directory for settings file if it doesn't exist
        settings_dir = os.path.dirname(self.settings_file)
        if not os.path.exists(settings_dir):
            os.makedirs(settings_dir)
        # dump settings to file
        with open(self.settings_file, "w") as fp:
            json.dump(self.settings, fp, indent=2)

    def set_status(self, message=None, msecs=None):
        """
        Display status of the database and other temporary messages in the status bar

        Parameters
        ----------
        message : str, optional
            Status message
        msecs : int, optional
            Duration of status message. By default the message will stay until overwritten.
        """
        # Update database status (permanent widget right of statusbar)
        self.db_status.setText("%d time series in database" % self.db.n)

        # show temporary message
        if not message:
            message = ""    # statusbar.showMessage() does not accept NoneType
        if not msecs:
            msecs = 0       # statusbar.showMessage() does not accept NoneType

        # self.statusBar().showMessage(message, msecs=msecs)

        # pyside6 uses 'timeout' keyword, pyqt6 uses 'msecs' keyword
        # solve pragmatically by try-except
        try:
            self.statusBar().showMessage(message, timeout=msecs)
        except TypeError:
            self.statusBar().showMessage(message, msecs=msecs)

    def selected_series(self):
        """
        Return list of names of checked series in item model
        """
        selected_items = []

        for row_number in range(self.db_proxy_model.rowCount()):
            # get index of item in proxy model and index of the same item in the source model
            proxy_index = self.db_proxy_model.index(row_number, 0)
            source_index = self.db_proxy_model.mapToSource(proxy_index)

            # is this item checked?
            is_selected = self.db_source_model.data(source_index, Qt.CheckStateRole) != 0

            if is_selected:
                # item path relative to common path in db
                rpath = self.db_source_model.data(source_index)

                # join with common path and add to list of checked items
                selected_items.append(os.path.join(self.db.common, rpath))

        return selected_items

    def start_times_series_processing_thread(self, container):
        """
        Start thread separate from main loop and process timeseries to calculate cycle distribution, power spectral
        density and max/min distributions.

        Parameters
        ----------
        container : dict
            Container with TimeSeries objects
        """
        self.set_status("Processing...", msecs=3000)

        # ui selections
        twin = self. time_window()
        fargs = self.filter_settings()
        nperseg = self.psd_nperseg()
        psdnorm = self.psd_normalized()
        nbins = self.rfc_nbins()
        minima_stats = self.minima.isChecked()

        # start calculation of filtered and windows time series trace
        worker = Worker(calculate_trace, container, twin, fargs)
        worker.signals.error.connect(self.log_thread_exception)
        worker.signals.result.connect(self.plot_trace)
        self.threadpool.start(worker)

        # start calculations of statistics
        worker = Worker(calculate_stats, container, twin, fargs, minima=minima_stats)
        worker.signals.error.connect(self.log_thread_exception)
        worker.signals.result.connect(self.tabulate_stats)
        worker.signals.result.connect(self.plot_weibull)
        self.threadpool.start(worker)

        # start calculations of psd
        worker = Worker(calculate_psd, container, twin, fargs, nperseg, psdnorm)
        worker.signals.error.connect(self.log_thread_exception)
        worker.signals.result.connect(self.plot_psd)
        self.threadpool.start(worker)

        # start calculations of rfc
        worker = Worker(calculate_rfc, container, twin, fargs, nbins)
        worker.signals.error.connect(self.log_thread_exception)
        worker.signals.result.connect(self.plot_rfc)
        self.threadpool.start(worker)

    def start_gumbel_fit_thread(self, container):
        """
        Start thread separate from main loop to sample extremes and fit Gumbel distribution to sample

        Parameters
        ----------
        container : dict
            Container with TimeSeries objects
        """
        self.set_status("Processing...", msecs=3000)

        # ui selections
        twin = self. time_window()
        fargs = self.filter_settings()

        # start calculation of filtered and windows time series trace
        worker = Worker(calculate_gumbel_fit, container, twin, fargs)
        worker.signals.error.connect(self.log_thread_exception)
        worker.signals.result.connect(self.plot_gumbel)
        self.threadpool.start(worker)

    def tabulate_stats(self, container):
        """
        Update table with time series statistics

        Parameters
        ----------
        container : dict
            Time series statistics
        """
        # disable sorting
        self.stats_table.setSortingEnabled(False)
        # populate table
        self.stats_table.setRowCount(max(len(container), 50))
        for i, (name, data) in enumerate(container.items()):
            for j, key in enumerate(STATS_ORDER):
                if key == "name":
                    cell = CustomTableWidgetItem(name)  # QTableWidgetItem(name)
                    cell.setToolTip(name)
                else:
                    value = data.get(key, np.nan)
                    cell = CustomTableWidgetItem(f"{value:12.5g}")   # works also with nan values
                cell.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
                self.stats_table.setItem(i, j, cell)
        # store original sorting order first time this function is called
        if self.stats_table_initial_sort is None:
            self.stats_table_initial_sort = self.stats_table.horizontalHeader().sortIndicatorSection()
        if self.stats_table_initial_order is None:
            self.stats_table_initial_order = self.stats_table.horizontalHeader().sortIndicatorOrder()
        # force original sorting order (in practice; same as in gui key list)
        self.stats_table.sortItems(self.stats_table_initial_sort, self.stats_table_initial_order)
        # re-enable sorting
        self.stats_table.setSortingEnabled(True)

    def time_window(self):
        """
        Time window from spin boxes
        """
        return self.from_time.value(), self.to_time.value()

    def update_model(self, newdb):
        """
        Fill item model with time series identifiers

        Parameters
        ----------
        newdb : TsDB
            Time series database
        """
        # merge the loaded time series into the database
        try:
            self.db.update(newdb)
        except KeyError:
            logging.error("The time series are not unique. You have probably loaded this file already.")
            return

        # fill item model with time series by unique id (common path is removed)
        names = self.db.list(names="*", relative=True, display=False)
        self.db_source_model.clear()    # clear before re-adding

        for name in names:
            # set each item as unchecked initially
            item = QStandardItem(name)
            item.setCheckState(Qt.Unchecked)
            item.setCheckable(True)
            item.setToolTip(os.path.join(self.db.common, name))
            item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsUserCheckable)
            self.db_source_model.appendRow(item)

        self.set_status()


class SettingsDialog(QDialog):
    """
    Dialog to configure application settings.

    Parameters
    ----------
    psdnorm : bool
        Default value for PSD normalization.
    nperseg : int
        Default number of points in segment when estimating PSD.
    nbins : int
        Default number of bins in cycle distribution.
    twindec : int
        Default number of decimals in time window for data processing.
    parent : QWidget, optional
        Parent widget.
    """
    def __init__(self, psdnorm, nperseg, nbins, twindec, parent=None):
        super(SettingsDialog, self).__init__(parent)
        self.setWindowTitle("Configure application settings")
        self.setWindowIcon(QIcon(ICON_FILE))
        layout = QVBoxLayout()

        # settings checkbox: normalized psd?
        self.psdnormcheckbox = QCheckBox()  # "Plot normalized power spectral density")
        self.psdnormcheckbox.setToolTip("Normalize power spectral density on maximum value to ease comparison of\n"
                                        "signals of different order of magnitude.")
        self.psdnormcheckbox.setChecked(False)
        if psdnorm:
            self.psdnormcheckbox.setChecked(True)
        # layout.addWidget(self.psdnormcheckbox)
        psdnormlayout = QHBoxLayout()
        psdnormlayout.addWidget(QLabel("Plot normalized power spectral density"))
        psdnormlayout.addStretch(1)
        psdnormlayout.addWidget(self.psdnormcheckbox)
        layout.addLayout(psdnormlayout)

        # settings spinbox: psd nperseg
        self.psdnpersegspinbox = QSpinBox()
        self.psdnpersegspinbox.setRange(100, 100000)
        self.psdnpersegspinbox.setSingleStep(10)
        self.psdnpersegspinbox.setEnabled(True)
        self.psdnpersegspinbox.setValue(nperseg)
        self.psdnpersegspinbox.setToolTip("When esimtating power spectral density using Welch's method the signal\n"
                                          "is dived into overlapping segments and psd is estimated for each segment\n"
                                          "and then averaged. The overlap is half of the segment length. The \n"
                                          "psd-estimate is smoother with shorter segments.")
        psdlayout = QHBoxLayout()
        psdlayout.addWidget(QLabel("Length of segment used when estimating power spectral density"))
        psdlayout.addStretch(1)
        psdlayout.addWidget(self.psdnpersegspinbox)
        layout.addLayout(psdlayout)

        # settings spinbox: rfc nbins
        self.rfcnbinsspinbox = QSpinBox()
        self.rfcnbinsspinbox.setRange(10, 1000)
        self.rfcnbinsspinbox.setSingleStep(1)
        self.rfcnbinsspinbox.setEnabled(True)
        self.rfcnbinsspinbox.setValue(nbins)
        self.rfcnbinsspinbox.setToolTip("Group the cycles counted using the Rainflow algorithm into a certain number\n"
                                        "of bins of equal width.")
        rfclayout = QHBoxLayout()
        rfclayout.addWidget(QLabel("Number of bins in cycle distribution based on RFC method"))
        rfclayout.addStretch(1)
        rfclayout.addWidget(self.rfcnbinsspinbox)
        layout.addLayout(rfclayout)

        # settings spinbox: time window number of decimals
        self.twindecspinbox = QSpinBox()
        self.twindecspinbox.setRange(1, 10)
        self.twindecspinbox.setSingleStep(1)
        self.twindecspinbox.setEnabled(True)
        self.twindecspinbox.setValue(twindec)
        self.twindecspinbox.setToolTip("Number of decimals in the data processing time window from/to boxes.")
        twindeclayout = QHBoxLayout()
        twindeclayout.addWidget(QLabel("Number of decimals in data processing time window *"))
        twindeclayout.addStretch(1)
        twindeclayout.addWidget(self.twindecspinbox)
        layout.addLayout(twindeclayout)

        # help text
        helptext = QHBoxLayout()
        helptext.addWidget(QLabel("* Close and re-open application for this setting to have effect"))
        helptext.addStretch(1)
        layout.addLayout(helptext)

        # buttons (OK, Cancel)
        self.buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal, self)
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)
        layout.addWidget(self.buttons)

        # conclude layout
        self.setLayout(layout)

    def get_settings(self):
        """Collect settings."""
        return self.psdnormcheckbox.isChecked(), self.psdnpersegspinbox.value(), self.rfcnbinsspinbox.value(), \
               self.twindecspinbox.value()

    @staticmethod
    def settings(defaults, parent=None):
        """Create settings dialog and return settings.

        Parameters
        ----------
        defaults : tuple
            Default values for parameters norm, nperseg and nbins.
        parent : QWidget, optional
            Parent widget

        Returns
        -------
        norm : bool
            Normalized PSD estimates.
        nperseg : int
            Number of points in segment when estimating PSD using Welch's method.
        nbins : int
            Number of bins in cycle distributions.
        twindec : int
            Number of decimals in time window for data processing.
        """
        dnorm, dnperseg, dnbins, dtwindec = defaults
        dialog = SettingsDialog(dnorm, dnperseg, dnbins, dtwindec, parent=parent)
        result = dialog.exec_()
        norm, nperseg, nbins, twindec = dialog.get_settings()
        # --- settings dialogue box debugging
        logging.debug(f"settings saved: norm = {norm}")
        logging.debug(f"settings saved: nperseg = {nperseg}")
        logging.debug(f"settings saved: nbins = {nbins}")
        logging.debug(f"settings saved: twindec = {twindec}")
        # ---
        return norm, nperseg, nbins, twindec, result == QDialog.Accepted





"""
Readers for SINTEF Ocean test data exhange format based on the Matlab .mat file.

Works for matlab file format version <=7.2 and >=7.3.
"""
import fnmatch
from datetime import datetime, timedelta
from typing import List, Tuple, Union

import numpy as np
from pymatreader import read_mat


def read_names(path):
    """
    Read time series names from SINTEF Ocean test data exhange format based on the Matlab .mat file.

    Parameters
    ----------
    path : str
        File path

    Returns
    -------
    str
        Name of the time array
    list
        Time series names

    Examples
    --------
    >>> tname, names = read_names('data.mat')

    """
    data = read_data(path)

    # identify time key, check that there is only one
    _tn = fnmatch.filter(data.keys(), '[Tt]ime*')
    if len(_tn) < 1:
        raise KeyError("File does not contain a time vector: %s" % path)
    elif len(_tn) > 1:
        raise KeyError("Duplicate time vectors on file: %s" % path)
    else:
        timename = _tn[0]

    # Keep only arrays of same size as time array
    tsize = data[timename].size
    names = [k for k, v in data.items() if (isinstance(v, np.ndarray) and v.size == tsize) and k != timename]

    return timename, names


def read_data(path: str, names: Union[List[str], Tuple[str]] = None):
    """
    Read time series data from SINTEF Ocean test data exhange format based on the Matlab .mat file.

    Parameters
    ----------
    path : str
        File path
    names : Union[List[str], Tuple[str]], optional
        Names of the requested time series incl. the time array itself. Defaults to all time series on the file.

    Returns
    -------
    dict
        Time and data

    Examples
    --------
    >>> tname, names = read_names('data.mat')
    >>> data = read_data('data.mat', [tname, *names])
    >>> t = data[tname]   # time
    >>> x1 = data[names[0]]  # first data series

    """
    # ignore the data field (if it exists) which contains the time series data in
    # latest file format
    data = read_mat(path)

    if "chan_names" in data.keys():
        # latest exhange format based on v.7.3 mat files
        data = dict(zip(data["chan_names"], np.transpose(data["data"])))
    else:
        # exhange format based on v.7.2 mat files
        ignored = ["comment", "fs", "test_num", "test_date", "__header__", "__version__", "__globals__"]
        data = {k: v for k, v in data.items() if k not in ignored}

    if names is not None:
        return {k: v for k, v in data.items() if k in names}
    else:
        return data


def _datenums_to_datetime(timearr):
    """
    Convert array of MATLAB datenum floats to array of datetime objects.

    Parameters
    ----------
    timearr: array_like or float
        Time array.

    Returns
    -------
    array
        Array of datetime objects, same shape as input array.
    """
    def convert(dn):
        # ref: https://stackoverflow.com/questions/13965740/converting-matlabs-datenum-format-to-python
        _dtg = datetime.fromordinal(int(dn)) + timedelta(days=dn % 1) - timedelta(days=366)
        return _dtg

    # return quickly if float (not array) is given
    if np.ndim(timearr) == 0:
        return convert(timearr)

    # convert array
    timearr = np.asarray(timearr)
    was_shape = timearr.shape
    timearr = timearr.flatten()
    dtarray = np.array([convert(t) for t in timearr])
    return dtarray.reshape(was_shape)




# -*- coding: utf-8 -*-
"""
Module for custom exceptions and exception handlers
"""
import os
import sys
import traceback
import webbrowser

from qtpy.QtWidgets import QMessageBox


def handle_exception(exc_type, exc_value, exc_traceback):
    """
    Handle all exceptions, log it on file, warn the user via a critical message box

    Parameters
    ----------
    exc_type : type
        type of exception e.g. ValueError
    exc_value : str
        exception string
    exc_traceback : traceback
        traceback for the exception

    """
    # retrieve traceback
    filename, line, dummy, dummy = traceback.extract_tb(exc_traceback).pop()
    filename = os.path.basename(filename)
    error = "%s: %s" % (exc_type.__name__, exc_value)

    # display critical message box
    msg = "<html>A critical error has occured.<br/> <b>%s</b><br/><br/>It occurred at <b>line %d</b> of file " \
        "<b>%s</b>.<br/>Press Yes to see the full error report<br/></html>" % (error, line, filename)
    message_box = QMessageBox().critical(None, "Error", msg.strip(), QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes)

    # pipe traceback to log file (env variable 'APPDATA' is assumed present on the computer)
    log_file = os.path.join(os.getenv("APPDATA"), "QATS.launch.pyw.log")
    with open(log_file,"w") as f:
        f.write("".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))

    # open log file in default editor if user pressed yes (trick using the webbrowser command)
    if message_box == QMessageBox.Yes:
        webbrowser.open(log_file)

    # exit with code 1 (error)
    sys.exit(1)


#!/usr/bin/env python
# coding: utf-8
"""
Transformations and operations related to motion.
"""
import numpy as np


def transform_motion(motion, newref, rotunit="deg"):
    """
    Transform motion to new reference position.

    The following sequence of rotation is used: z-y-z (known as yaw-pitch-roll). For more detailed description,
    see https://en.wikipedia.org/wiki/Euler_angles or SIMO Theory Manual (ch. 6.2 in version 4.14.0).

    Parameters
    ----------
    motion : tuple or np.ndarray
        Motion of old reference point, 6 dofs (hence shape (6, nt), where nt is number of time steps):
            - x (position of old reference point in global coord. system)
            - y (position of old reference point in global coord. system)
            - z (position of old reference point in global coord. system)
            - rx (rotation about local x-axis, aka. roll)
            - ry (rotation about local x-axis, aka. roll)
            - rz (rotation about local x-axis, aka. roll)
    newref : 3-tuple
        Position vector (in body coordinate system) of new reference point: (x, y, z).
    rotunit : {'deg', 'rad'}
        Unit of rotations. Default is degrees.

    Returns
    -------
    xyz : np.ndarray
        Position vector for new reference position, shape (3, n).
    """
    motion = np.asarray(motion)  # ensure motion is numpy array
    newref = np.asarray(newref)  # ensure newref is numpy array, for efficiency in loop with np.dot
    ndof, nt = motion.shape      # number of dofs and time steps
    assert ndof == 6, f"Motion must be of shape (6, nt) (6-dof motion), got {motion.shape}"
    assert newref.size == 3, f"Specified position must be list/tuple with three values, got {len(newref)}"

    # extract rotations, scale to radians if needed given as degrees
    if rotunit == "deg":
        rx, ry, rz = np.radians(motion[-3:, :])
    elif rotunit == "rad":
        rx, ry, rz = motion[-3:, :]
    else:
        raise ValueError(f"Parameter `rotunit` must be 'deg' or 'rad', not '{rotunit}'")

    # calculate transformation matrix -> shape (3, 3, nt)
    trans = np.array([
        [np.cos(rz) * np.cos(ry), -np.sin(rz) * np.cos(rx) + np.cos(rz) * np.sin(ry) * np.sin(rx),
         np.sin(rz) * np.sin(rx) + np.cos(rz) * np.sin(ry) * np.cos(rx)],
        [np.sin(rz) * np.cos(ry), np.cos(rz) * np.cos(rx) + np.sin(rz) * np.sin(ry) * np.sin(rx),
         -np.cos(rz) * np.sin(rx) + np.sin(rz) * np.sin(ry) * np.cos(rx)],
        [-np.sin(ry), np.cos(ry) * np.sin(rx), np.cos(ry) * np.cos(rx)],
    ])

    # transform to new reference position for each time step -> shape (3, nt)
    xyz = np.zeros((3, nt))
    for i in range(nt):
        xyz[:, i] = np.dot(trans[:, :, i], newref)
    # ... and add xyz motion of old reference point
    xyz += motion[:3, :]

    return xyz


def acceleration(x, t):
    """
    Numerical time differentiation to obtain acceleration of signal(s).


    Parameters
    ----------
    x : list or np.array
        Signal(s) to perform time derivation on. See requirements given for function :func:`velocity`.
    t : float or np.array
        Time step or time array.

    Returns
    -------
    acc : np.array
        Acceleration of input signal(s); same shape as input.

    Notes
    -----
    In practice, acceleration is calculated by calling function `velocity` twice.

    See Also
    --------
    velocity
    numpy.gradient
    """
    acc = x[:]  # copy input signals
    for _ in range(2):
        acc = velocity(acc, t)
    return acc


def velocity(x, t):
    """
    Numerical time differentiation to obtain velocity of signal(s).

    Parameters
    ----------
    x : list or np.array
        Signal to perform time derivation on. If a 2-D array is given, it is required to be of shape (n, nt), where `n`
        is number of signals and `nt` is number of time steps. Time differentiation is then performed for each of the
        signals independently.
    t : float or np.array
        Time step or time array.

    Returns
    -------
    vel : np.array
        Velocity of input signal(s); same shape as input.

    Notes
    -----
    Velocity is calculated using second order central differences, with first order backwards and forward differences
    at start/end of signals.
    See `numpy.gradient <https://docs.scipy.org/doc/numpy/reference/generated/numpy.gradient.html>`_.


    See Also
    --------
    numpy.gradient

    """
    # check input signal(s)
    x = np.asarray(x)
    assert x.ndim in (1, 2), "Input signal 'x' must be 1- or 2-D array"
    # check specified time
    if isinstance(t, float):
        # time step is specified
        pass
    else:
        # assume time array is specified -> check that it matches the number of time steps in input signals
        t = np.asarray(t)   # ensure t is numpy array
        nt = x.shape[-1]    # number of time steps
        assert t.size == nt, "If time array is specified, it must match number of time steps in input signal(s)"

    if x.ndim == 1:
        axis = None
    else:
        # more than one signal is given -> perform time differentiation along axis 1
        axis = 1

    # time differentiation
    vel = np.gradient(x, t, axis=axis)

    return vel




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module implementing a logger class

@author: perl
"""
import logging

from qtpy.QtWidgets import QTextBrowser


class QLogger(logging.Handler):
    """
    Logger handler
    """
    def __init__(self, parent):
        super().__init__()
        self.widget = QTextBrowser(parent)
        self.widget.setReadOnly(True)

    def clear(self):
        """Clear widget."""
        self.widget.setText("")

    def emit(self, record):
        """
        Append logger record to text browser.

        Parameters
        ----------
        record : str
            logger record

        """
        record = self.format(record)
        self.widget.append(record)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:class:`Gumbel` class and functions related to Gumbel distribution.
"""
import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import fsolve, leastsq
from scipy.special import binom, zetac

from .empirical import empirical_cdf

# todo: build documentation and check that docstrings behave as intended
# todo: create examples
# todo: consider replacing latex with python code or pseudo code (ref. link below)
# link: https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt#other-points-to-keep-in-mind


class Gumbel(object):
    """
    The Gumbel maxima distribution.

    The cumulative distribution function is defined as::

        F(x) = exp{-exp[-(x-a)/b]}

    where `a` is location parameter and `b` is the scale parameter.

    Parameters
    ----------
    loc: float
        Gumbel location parameter.
    scale: float
        Gumbel scale parameter.
    data: array_like, optional
        Sample data, used to establish empirical cdf and is included in plots.
        To fit the Gumbel distribution to the sample data, use :meth:`Gumbel.fit`.

    Attributes
    ----------
    loc : float
        Gumbel location parameter.
    scale : float
        Gumbel scale parameter.
    data : array_like
        Sample data.

    Examples
    --------
    To initiate an instance based on parameters, use:

    >>> from qats.stats.gumbel import Gumbel
    >>> gumb = Gumbel(loc, scale)

    If you need to establish a Gumbel instance based on a sample data set, use:

    >>> gumb = Gumbel.fit(data, method='msm')


    References
    ----------
    1. Statistical models in applied science., Bury, K.V. (1975), Wiley, New York

    2. Bruk av asymptotiske ekstremverdifordelinger, Haver, S. (2007)

    3. `Plotting positions <http://en.wikipedia.org/wiki/Q%E2%80%93Q_plot>`_, About plotting positions

    4. `Usable estimators for parameters in Gumbel distribution
       <http://stats.stackexchange.com/questions/71197/usable-estimators-for-parameters-in-gumbel-distribution>`_

    5. `Bootstrapping statistics <https://en.wikipedia.org/wiki/Bootstrapping_(statistics)>`_

    6. Probability weighted moments, Greenwood, J. A.; Landwehr, J.M.; Matalas, N.C.; Wallis, J.R., 1979,
       Water Resources Research. 15(5): 1049-1054.

    7. Probability weighted moments compared with some traditional techniques in estimating gumbel parameters and
       quantiles., Landwehr, J.M.; Matalas, N.C.; Wallis, J.R., 1979., Water Resources Research. 15(5): 1063-1064.

    """

    def __init__(self, loc, scale, data=None):
        assert loc is not None, "Location parameter must be finite."
        assert scale is not None and scale > 0, "Scale parameter must be finit and larger than 0."
        self.loc = loc
        self.scale = scale

        if data is not None:
            self.data = np.array(data)
        else:
            self.data = None

    @property
    def cov(self):
        """
        Distribution coefficient of variation (C.O.V.)

        Returns
        -------
        float
            Distribution c.o.v.
        """
        return self.std / self.mean

    @property
    def ecdf(self):
        """
        Median rank empirical cumulative distribution function associated with the sample

        Returns
        -------
        array
            Empirical cumulative distribution function

        Notes
        -----
        Requires data/sample to be specified.

        Gumbel recommended the following mean rank quantile formulation ``Pi = i/(n+1)``.
        This formulation produces a symmetrical CDF in the sense that the
        same plotting positions will result from the data regardless of
        whether they are assembled in ascending or descending order.

        A more sophisticated median rank formulation ``Pi = (i-0.3)/(n+0.4)`` approximates the
        median of the distribution free estimate of the sample variate to about
        0.1% and, even for small values of n, produces parameter estimations
        comparable to the result obtained by maximum likelihood estimations (Bury, 1999, p43).
        A median rank method, ``pi=(i-0.3)/(n+0.4)``, is chosen to approximate the mean
        of the distribution [2].

        The empirical cdf is also used as plotting positions when plotting the sample
        on probability paper.
        """
        assert self.data is not None, "Requires data/sample to be specified."
        return empirical_cdf(self.data.size, kind='median')

    @property
    def kurt(self):
        """
        Distribution kurtosis

        Returns
        -------
        float
            Distribution kurtosis
        """
        return 12. / 5.

    @property
    def mean(self):
        """
        Distribution mean value

        Returns
        -------
        float
            Distribution mean value
        """
        return self.loc + self.scale * _euler_masceroni()

    @property
    def median(self):
        """
        Distribution median value

        Returns
        -------
        float
            Distribution median value
        """
        return self.loc - self.scale * np.log(np.log(2.))

    @property
    def mode(self):
        """
        Distribution mode value

        Returns
        -------
        float
            Distribution mode value
        """
        return self.loc

    @property
    def mse(self):
        """
        Mean squared error of fitted cumulative distribution (a,b,c) and empirical distribution

        Returns
        -------
        float
            mean squared error

        Notes
        -----
        Requires data/sample to be specified.

        """
        return np.sum((self.ecdf - self.cdf(np.sort(self.data))) ** 2.) / self.data.size

    @property
    def params(self):
        """
        Distribution parameters.

        Returns
        -------
        tuple
            Distribution parameters: (loc, scale).
        """
        return self.loc, self.scale

    @property
    def std(self):
        """
        Distribution standard deviation

        Returns
        -------
        float
            Distribution standard deviation
        """
        return np.pi * self.scale / np.sqrt(6)

    @property
    def skew(self):
        """
        Distribution skewness

        Returns
        -------
        float
            Distribution skewness

        Notes
        -----
        `zetac` is the complementary Riemann zeta function (zeta function minus 1).
        See http://docs.scipy.org/doc/scipy/reference/generated/scipy.special.zetac.html

        """
        return 12. * np.sqrt(6.) * (1. + zetac(3)) / np.pi ** 3

    def cdf(self, x=None):
        """
        Cumulative distribution function (cumulative probability) for specified values x

        Parameters
        ----------
        x : array_like, optional
            Calculate cumulative probability for these values

        Returns
        -------
        array
            Cumulative probabilities for specified values x

        Notes
        -----
        A range of x values [loc, loc+3*std] are applied if x is not specified.

        """
        assert self.scale > 0., "The scale parameter must be larger than 0."

        if x is None:
            x = np.linspace(self.loc, self.loc + 3. * self.std, 100)
        else:
            x = np.array(x)

        z = (x - self.loc) / self.scale
        p = np.exp(-np.exp(-z))
        return p

    @classmethod
    def fit(cls, data, method='msm', verbose=False):
        """
        Determine distribution parameters by fit to sample.

        Parameters
        ----------
        data : array_like
            Sample
        method : str, optional
            Method of fit. Options:

            - ``msm`` = method of sample moments (default)
            - ``lse`` = least-square estimation
            - ``mle`` = maximum likelihood estimation
            - ``pwm`` = probability weighted moments

        verbose : bool, optional
            If true, fitted parameters are written to screen.

        Examples
        --------
        Assuming `data` is a sample array/list:

        >>> from qats.stats.gumbel import Gumbel
        >>> gumb = Gumbel.fit(data, method="msm")

        """
        options = {'msm': msm, 'lse': lse, 'pwm': pwm, 'mle': mle}
        assert method.lower() in options.keys(), "Method must be either %s" % (' or '.join(options.keys()))

        data = np.array(data)  # ensure numpy array

        loc, scale = options[method](data)
        if verbose:
            print("Fitted parameters:\nloc = %5.3g\nscale = %5.3g" % (loc, scale))

        gumb = cls(loc, scale, data=data)

        return gumb

    @classmethod
    def fit_from_weibull_parameters(cls, wa, wb, wc, n, verbose=False):
        """
        Calculate Gumbel distribution parameters from n independent Weibull distributed variables.

        Parameters
        ----------
        wa : float
            Weibull loc parameter
        wb : float
            Weibull scale parameter
        wc : float
            Weibull shape parameter
        n : int
            Number independently distributed variables
        verbose : bool
            Print fitted parameters

        Notes
        -----
        A warning is issued if Weibull shape parameter less than 1. In this case,
        the convergence towards asymptotic extreme value distribution is slow
        , and the asymptotic distribution will be non-conservative relative
        to the exact distribution. The asymptotic distribution is correct with Weibull
        shape equal to 1 and conservative with Weibull shape larger than 1.
        These deviations diminish with larger samples. See [1, p. 380].

        References
        ----------
        1. Bury, Karl V., 1975, "Statistical Models in Applied Science", University of British Columbia, John Wiley & Sons

        Examples
        --------
        >>> from qats.stats.gumbel import Gumbel
        >>> gumb = Gumbel.fit_from_weibull_parameters(wa, wb, wc, n)
        """

        loc = wa + wb * np.log(n) ** (1. / wc)  # eq. 11.41 on page 380 in [1]
        scale = (wb / wc) * np.log(n) ** ((1.-wc) / wc)  # eq. 11.41 on page 380 in [1]

        if verbose:
            print("Fitted parameters:\nloc = %5.3g\nscale = %5.3g" % (loc, scale))

        gumb = cls(loc, scale)

        return gumb

    def invcdf(self, p=None):
        """
        Inverse cumulative distribution function for specified probabilities

        Parameters
        ----------
        p : array_like, optional
            Calculate the inverse cumulative distribution function for these probabilities

        Returns
        -------
        array
            Values corresponding to the specified quantiles

        Notes
        -----
        A range of quantiles from 0.001 to 0.999 are applied if quantiles are not specified

        """
        assert self.scale > 0., "The scale parameter must be larger than 0."

        if p is None:
            p = np.linspace(0.001, 0.999, 100)
        else:
            p = np.array(p)

        x = np.zeros(np.shape(p))

        x[p == 1.] = np.inf  # asymptotic
        x[(p < 0.) | (p > 1.)] = np.nan  # probabilities out of bounds
        z = (p >= 0.) & (p < 1.)  # valid quantile range
        x[z] = self.loc - self.scale * np.log(-np.log(p[z]))

        return x

    def pdf(self, x=None):
        """
        Probability density function for specified values x

        Parameters
        ----------
        x : array_like, optional
            Cumulative probabilities for specified values x

        Returns
        -------
        array
            Calculate probability density for these values x

        Notes
        -----
        A range of x values [loc, loc+3*std] are applied if x is not specified.

        """
        assert self.scale > 0., "The scale parameter must be larger than 0."

        if x is None:
            x = np.linspace(self.loc, self.loc + 3. * self.std, 100)
        else:
            x = np.array(x)

        z = (x - self.loc) / self.scale
        p = (1. / self.scale) * np.exp(-z - np.exp(-z))
        return p

    def plot(self, filename=None):
        """
        Plot cumulative distribution function

        Parameters
        ----------
        filename : str, optional
            Save plot as `filename`, default is to show plot on screen

        """
        plt.figure()

        # plot sample versus empirical distribution
        if self.data is not None:
            plt.plot(np.sort(self.data), self.ecdf, 'ko', label='Data')

        # plot fitted/specified distribution
        y = np.linspace(0.001, 0.9999, 1000)
        x = self.invcdf(p=y)
        plt.plot(x, y, '-r', label='Fitted')

        plt.xlabel('X')
        plt.ylabel('Cumulative probability')
        plt.legend(loc='upper left')
        plt.grid(True)

        if filename is not None:
            plt.savefig(filename)

        else:
            plt.show()

        # close figure to avoid high memory consumption
        plt.close()

    def plot_linear(self, filename=None):
        """
        Plot cumulative distribution function on linearized Gumbel scales

        Parameters
        ----------
        filename : str, optional
            Save plot as `filename`, default is to show plot on screen

        """
        plt.figure()

        # plot sample versus empirical distribution function
        if self.data is not None:
            x = np.sort(self.data)
            z = -np.log(-np.log(self.ecdf))
            plt.plot(x, z, 'ko', label='Data')

        # plot distribution
        x = self.invcdf(p=np.linspace(0.001, 0.9999, 1000))
        z = (x - self.loc) / self.scale
        plt.plot(x, z, '-r', label='Fitted')

        # plotting positions and plot configurations
        p = np.array([0.1, 0.2, 0.5, 0.7, 0.8, 0.9, 0.95, 0.99, 0.999, 0.9999])
        y = -np.log(-np.log(p))
        plt.yticks(y, p)
        plt.legend(loc='upper left')
        plt.ylim(y[0], y[-1])
        plt.xlabel('X')
        plt.ylabel('Cumulative probability')
        plt.grid(True)

        if filename is not None:
            plt.savefig(filename)
        else:
            plt.show()

        # close figure to avoid high memory consumption
        plt.close()

    def rnd(self, size=None, seed=None):
        """
        Draw random samples from probability distribution

        Parameters
        ----------
        size : int|numpy shape, optional
            Sample size (default 1 random value is returned)
        seed : int, optional
            Seed for random number generator (default seed is random)

        Returns
        -------
        array
            Random sample

        Examples
        --------
        Pick 1000 values randomly from a Gumbel distribution

        >>> from qats.stats.gumbel import Gumbel
        >>> g = Gumbel(loc, scale)
        >>> sample = g.rnd(size=1000)

        If you want to preset the seed for the random sampling (to be able to repeat the sampling)

        >>> from qats.stats.gumbel import Gumbel
        >>> g = Gumbel(loc, scale)
        >>> sample = g.rnd(size=1000, seed=3)

        """
        if seed is not None:
            np.random.seed(seed)
        r = np.random.random_sample(size)
        x = self.invcdf(p=r)

        return x


def _euler_masceroni():
    """
    The Euler-Masceroni constant

    Returns
    -------
    float
        The Euler-Mascheroni constant

    Notes
    -----
    The Euler–Masceroni constant (also called Euler's constant) is a
    mathematical constant recurring in analysis and number theory. It
    is defined as the limiting difference between the harmonic series and
    the natural logarithm.

    """
    return 0.57721566490153286060651209008240243104215933593992


def bootstrap(loc, scale, size, repetitions, method='pwm'):
    """
    Quantify mean and coefficient of variation of Gumbel distribution parameters using parametric bootstrapping

    Parameters
    ----------
    loc : float
        Source distribution location parameter
    scale : float
        Source distribution scale parameter
    size : int
        Size of bootstrapped sample
    method : str, optional
        method of fit, optional
        'msm' = method of sample moments
        'lse' = least-square estimation
        'mle' = maximum likelihood estimation
        'pwm' = probability weighted moments (default)
    repetitions : int, optional
        Number of bootstrap samples. default equal to 100

    Returns
    -------
    array
        Mean distribution parameters
    array
        Coefficient of variation of distribution parameter

    Notes
    -----
    In statistics, bootstrapping is a method for assigning measures of accuracy
    to sample estimates (variance, quantiles). This technique allows estimation of the
    sampling distribution  of almost any statistic using only very simple methods. Generally,
    it falls in the broader class of resampling methods. In this case a parametric model is fitted
    to the data, and samples of random numbers with the same size as the original data,
    are drawn from this fitted model. Then the quantity, or estimate, of interest is
    calculated from these data. This sampling process is repeated many times as for other
    bootstrap methods. If the results really matter, as many samples as is reasonable,
    given available computing power and time, should be used. Increasing the number of
    samples cannot increase the amount of information in the original data, it can only
    reduce the effects of random sampling errors which can arise from a bootstrap procedure itself.
    See [5] about bootstrapping.

    Examples
    --------
    To quantify the uncertainty (coefficient of variation) of a Gumbel distribution fitted to a sample with 5 values
    (using 100 repetition):

    >>> from qats.stats.gumbel import bootstrap
    >>> m, cv = bootstrap(10, 2.5, 5, 100)

    """
    options = {'msm': msm, 'lse': lse, 'pwm': pwm, 'mle': mle}
    assert method.lower() in options.keys(), "Method must be either %s" % (' or '.join(options.keys()))

    # initiate distribution
    distribution = Gumbel(loc, scale)

    i = 0
    par = np.zeros((repetitions, 2))
    while i < repetitions:
        # sample random numbers from source distribution
        x = distribution.rnd(size=size)

        # estimate distribution parameters to random sample
        par[i, :] = options[method](x)
        i += 1

    # calculate mean value and coefficient of variation of the distribution parameters
    m = par.mean(axis=0)
    cv = par.std(axis=0, ddof=1) / m

    return m, cv


def lse(x):
    """
    Fit Gumbel distribution parameters to sample by method of least square fit to empirical cdf

    Parameters
    ----------
    x : array_like
        data sample

    Returns
    -------
    floats
        distribution `loc` and `scale` parameters

    Notes
    -----
    Uses an approximate median rank estimate for the empirical cdf.
    """
    x = np.sort(x)
    f = empirical_cdf(x.size, kind='median')  # median rank cdf
    fp = lambda v, z: np.exp(-np.exp(-(z - v[0]) / v[1]))  # parametric Gumbel function
    e = lambda v, z, y: (fp(v, z) - y)  # error function to be minimized
    a0, b0 = msm(x)  # initial guess based on method of moments

    # least square fit
    p, cov, info, msg, ier = leastsq(e, np.array([a0, b0]), args=(x, f), full_output=True)

    return p[0], p[1]


def mle(x):
    """
    Fit distribution parameters to sample by maximum likelihood estimation

    Parameters
    ----------
    x : array_like
        data sample

    Returns
    -------
    floats
        distribution `loc` and `scale` parameters

    Notes
    -----
    MLE equation set is given in 'Statistical Distributions' by Forbes et.al. (2010) and referred
    at [4]
    """

    def mle_eq(p, z):
        """
        MLE equation set

        Parameters
        ----------
        p : list
            distribution parameters a, b, c
        z : array_like
            data sample
        """
        loc, scale = p  # unpack parameters
        n = z.size

        out = [loc + scale * np.log(1. / n * np.sum(np.exp(-z / scale))),
               z.mean() - np.sum(z * np.exp(-z / scale)) / np.sum(np.exp(-z / scale)) - scale]

        return out

    x = np.array(x)
    a0, b0 = msm(x)  # initial guess
    a, b = fsolve(mle_eq, np.array([a0, b0]), args=x)

    return a, b


def msm(x):
    """
    Fit Gumbel distribution parameters to sample by method of sample moments

    Parameters
    ----------
    x : array_like
        data sample

    Returns
    -------
    floats
        distribution `loc` and `scale` parameters

    Notes
    -----
    See description in [1] and [2].
    """
    x = np.array(x)

    b = np.sqrt(6.) * np.std(x, ddof=1) / np.pi  # using the unbiased sample standard deviation
    a = x.mean() - _euler_masceroni() * b

    return a, b


def plot_fits(data, filename=None, methods=None):
    """
    Plot data sample versus empirical and fitted cumulative distribution function on linearized Gumbel scales

    Parameters
    ----------
    data : array_like
        Data sample
    filename : str, optional
        Save plot as `filename`, default is to show plot on sc
    methods : tuple, optional
        Methods of fit. Options (default all):

        - ``msm`` = method of sample moments
        - ``lse`` = least-square estimation
        - ``mle`` = maximum likelihood estimation
        - ``pwm`` = probability weighted moments
    """
    plt.figure()

    # sort data, create empirical distribution function and plot
    x = np.sort(data)
    z = -np.log(-np.log(empirical_cdf(x.size, kind='median')))
    plt.plot(x, z, 'ko', label='Data')

    # fit distributions and plot
    options = {'msm': msm, 'lse': lse, 'pwm': pwm, 'mle': mle}

    # default if not specified
    if not methods:
        methods = options.keys()

    for method in methods:
        assert method.lower() in options.keys(), "Method must be either %s" % (' or '.join(options.keys()))

        # estimate location and scale parameter
        loc, scale = options[method](x)

        # linarize and plot
        z = (x - loc) / scale
        plt.plot(x, z, label=method)

    # adjust plotting positions and plot configurations
    p = np.array([0.1, 0.2, 0.5, 0.7, 0.8, 0.9, 0.95, 0.99, 0.999, 0.9999])
    y = -np.log(-np.log(p))
    plt.yticks(y, p)
    plt.legend(loc='upper left')
    plt.ylim(y[0], y[-1])
    plt.xlabel('data')
    plt.ylabel('Cumulative probability')
    plt.grid(True)

    if filename is not None:
        plt.savefig(filename)
    else:
        plt.show()

    # close figure to avoid high memory consumption
    plt.close()


def pwm(x):
    """
    Fit Gumbel distribution parameters to sample by method of probability weighted
    moments [7].

    Parameters
    ----------
    x : array_like
        data sample

    Returns
    -------
    tuple
        distribution parameters `location` and `scale`

    """

    def mk(z, k):
        """
        Probability weighted moments with emphasize on the left/lower tail (j=0).

        Parameters
        ----------
        z   :   array
            sample
        k   :   order of complementary cumulative distribution function

        Notes
        -----
        The probability weighted moment Mljk is defined by Greenwood and others (1979)

        .. math::

            M_{l,j,k} = E[X^l F^j (1-F)^k]

        , where X(F) is the inverse form of the distribution and F is the cumulative distribution function.
        When j=k=0 and l is a non-negative integer then M_{l,0,0} represents the conventional moment of order l about
        the origin.

        PWMs can be applied either when the small observations are more important than the large observations (k=0), as in
        strength properties of materials, or when the large observations should have more influence than the smaller
        observations (k=0) as with three diameter distribution modelling. Here we have chose the former and derived
        unbiased estimators for moments M_{1,0,k} (j=0), see eq. 16 in [6].

        .. math::

            M_{1,0,k} = \frac{1}{n}\sum_{i=1}^{n}{X_{(i)}\frac{\binom{n-i}{k}}{\binom{n-1}{k}}}

        """
        n = float(z.size)
        xi = np.sort(z)
        ii = np.arange(1, n + 1)

        # ratio of binomial coefficients
        bc = np.array([binom(n - i, k) / binom(n - 1., k) for i in ii])

        return (1. / n) * np.sum(xi * bc)  # eq. 16 in [6]

    m0 = mk(x, 0.)
    m1 = mk(x, 1.)

    b = (m0 - 2. * m1) / np.log(2.)
    a = m0 - _euler_masceroni() * b

    return a, b




"""
Readers CSV formatted time series files
"""
import pandas as pd


def read_names(path):
    """
    Read time series names from a comma-separated file.

    Parameters
    ----------
    path : str
        Name of .csv file

    Returns
    -------
    list
        Time series names

    Notes
    -----
    The series names are expected found on the header row. Time is expected to be in the first column.

    """
    # pandas will infer the format e.g. delimiter.
    df = pd.read_csv(path, nrows=1, sep=None, engine='python', encoding='utf-8')
    names = list(df)
    _ = names.pop(0)    # remove time which is assumed to be in the first column
    return names


def read_data(path, ind=None):
    """
    Read time series data arranged column wise on a comma-separated file.

    Parameters
    ----------
    path : str
        CSV file path (relative or absolute)
    ind : list|tuple, optional
        Read only a subset of the data series specified by index , with 0 being the first. For example,
        ind = (1,4,5) will extract the 2nd, 5th and 6th columns. The default, None, results in all columns being read.

    Returns
    -------
    array
        Time and data

    """
    df = pd.read_csv(path, usecols=ind, sep=None, engine='python')  # pandas will infer the format e.g. delimiter.
    return df.T.to_numpy()



"""
Module with functions for handling file operations and calculations. Made for multithreading.
"""
import numpy as np

from ..fatigue.rainflow import rebin as rebin_cycles
from ..stats.gumbel import pwm as gumbel_pwm
from ..tsdb import TsDB


def calculate_psd(container, twin, fargs, nperseg, normalize):
    """
    Calculate power spectral density and return as numpy arrays

    Parameters
    ----------
    container : dict
        TimeSeries objects
    twin : tuple
        Time window. Time series are cropped to time window before estimating psd.
    fargs : tuple
        Filter arguments. Time series are filtered before estimating psd.
    nperseg : int
        Size of segments used for ustimating PSD using Welch's method.
    normalize : bool
        Normalize power spectral density on maximum density.

    Returns
    -------
    dict
        Frequency versus power spectral density
    """
    container_out = dict()

    for name, ts in container.items():
        # resampling to average time step for robustness (necessary for series with varying time step)
        f, s = ts.psd(twin=twin, filterargs=fargs, resample=ts.dt, taperfrac=0.1, nperseg=nperseg, normalize=normalize)
        container_out[name] = tuple([f, s])

    return container_out


def calculate_rfc(container, twin, fargs, nbins):
    """
    Count cycles using Rainflow counting method

    Parameters
    ----------
    container : dict
        TimeSeries objects
    twin : tuple
        Time window. Time series are cropped to time window before estimating psd.
    fargs : tuple
        Filter arguments. Time series are filtered before estimating psd.
    nbins : int
        Number of bins in cycle distribution.

    Returns
    -------
    dict
        Cycle magnitude versus count
    """
    container_out = dict()

    for name, ts in container.items():
        # rainflow counting
        cycles = ts.rfc(twin=twin, filterargs=fargs)

        # rebin
        if nbins is not None:
            cycles = rebin_cycles(cycles, binby='range', n=nbins)

        # unpack pairs (tuples) in list
        r, _, c = zip(*cycles)
        container_out[name] = tuple([r, c])

    return container_out


def calculate_trace(container, twin, fargs):
    """
    Calculate trace and peaks/troughs of filtered time series

    Parameters
    ----------
    container : dict
        TimeSeries objects
    twin : tuple
        Time window. Time series are cropped to time window before extracting maxima.
    fargs : tuple
        Filter arguments. Time series are filtered before extracting maxima.

    Returns
    -------
    dict
        Filtered and windowed time series and peaks/throughs
    """
    container_out = dict()

    for name, ts in container.items():
        t, x = ts.get(twin=twin, filterargs=fargs)
        xmin, tmin = ts.minima(twin=twin, filterargs=fargs, rettime=True)
        xmax, tmax = ts.maxima(twin=twin, filterargs=fargs, rettime=True)

        container_out[name] = dict(t=t, x=x, tmin=tmin, xmin=xmin,
                                   tmax=tmax, xmax=xmax)

    return container_out


def calculate_stats(container, twin, fargs, minima=False):
    """
    Calculate time series statistics

    Parameters
    ----------
    container : dict
        TimeSeries objects
    twin : tuple
        Time window. Time series are cropped to time window before extracting maxima.
    fargs : tuple
        Filter arguments. Time series are filtered before extracting maxima.
    minima : bool, optional
        Fit to sample of minima instead of maxima. The sample is multiplied by -1 prior to parameter estimation.

    Returns
    -------
    dict
        Filtered and windowed time series and peaks/throughs
    """
    return {name: ts.stats(twin=twin, filterargs=fargs, statsdur=10800., quantiles=(0.37, 0.57, 0.9),
                           is_minima=minima, include_sample=True) for name, ts in container.items()}


def calculate_gumbel_fit(container, twin, fargs, minima=False):
    """
    Calculate time series extremes and fit Gumbel distribution

    Parameters
    ----------
    container : dict
        TimeSeries objects
    twin : tuple
        Time window. Time series are cropped to time window before extracting maxima.
    fargs : tuple
        Filter arguments. Time series are filtered before extracting maxima.
    minima : bool, optional
        Fit to sample of minima instead of maxima. The sample is multiplied by -1 prior to parameter estimation.

    Returns
    -------
    dict
        Sample and fitted gumbel distribution parameters
    """
    if len(container.keys()) < 2:
        raise ValueError(f"Select more than 1 time series to fit Gumbel CDF to extremes sample.")

    # create sample of extremes
    sample = list()
    for name, ts in container.items():
        _, x = ts.get(twin=twin, filterargs=fargs)

        if minima:
            # sample minimum value
            sample.append(np.min(x))
        else:
            # sample maximum value
            sample.append(np.max(x))

    # numpy array sorted ascending
    sample = np.sort(np.array(sample))

    # flip sample of minima to enable Gumbel fit
    if minima:
        sample *= -1

    # estimate gumbel distribution parameters
    loc, scale = gumbel_pwm(sample)
    return dict(loc=loc, scale=scale, sample=sample, minima=minima)


def export_to_file(filename, db, names, twin, fargs):
    """
    Export selected time series to file

    Parameters
    ----------
    filename : str
        File name
    db : TsDB
        Time series data base
    names : list
        Name of time series to export
    twin : tuple
        Time window. Time series are cropped to time window before export.
    fargs : tuple
        Filter arguments. Time series are filtered before export
    """
    db.export(filename, names=names, exist_ok=True, basename=False, twin=twin, filterargs=fargs)


def import_from_file(files):
    """
    Import time series files

    Parameters
    ----------
    files : list
        File names

    Returns
    -------
    TsDB
        Time series data base
    """
    db = TsDB()
    db.load(files, read=False)
    return db


def read_timeseries(db, names):
    """
    Read time series from files

    Parameters
    ----------
    db : TsDB
        Time series data base
    names : list
        Name of time series to read

    Returns
    -------
    dict
        Container with TimeSeries objects
    """
    return db.getm(names=names, store=False)




# -*- coding: utf-8 -*-
"""
Command line interface to app (GUI).
"""
import argparse
import os
import sys

from pkg_resources import resource_filename
from qtpy.QtWidgets import QApplication

from . import __version__
from .app.exceptions import handle_exception
from .app.gui import LOGGING_LEVELS, Qats


def link_app():
    """
    Create start menu item and desktop shortcut to `qats` desktop app
    (by invoking pythonw.exe with the necessary arguments, not the 
    previous entry point qats-app.exe (or .cmd)).
    """
    if not sys.platform == "win32":
        print(f"Unable to create links to app on {sys.platform} OS.")
        sys.exit()

    from win32com.client import Dispatch

    pkg_name = "qats"
    ico_path = resource_filename("qats.app", "qats.ico")
    # target = os.path.join(scripts_dir, f"{pkg_name}-app.exe")
    lnk_name = pkg_name.upper() + ".lnk"

    # define target as pythonw.exe (or python.exe if needed)
    python_exec_path = sys.executable
    python_exec_dir = os.path.dirname(python_exec_path)
    pythonw_exec_path = os.path.join(python_exec_dir, 'pythonw.exe')
    if os.path.exists(pythonw_exec_path):
        target = pythonw_exec_path
    else:
        target = python_exec_path

    # define arguments to target
    # (relies on invoking qats.__main__, not the entry point executable)
    arguments = f"-m {pkg_name} app"
    
    # open shell
    shell = Dispatch("WScript.Shell")

    # create shortcuts to gui in desktop folder and start-menu programs
    for loc in ("Desktop", "Programs"):
        location = shell.SpecialFolders(loc)
        path_link = os.path.join(location, lnk_name)
        shortcut = shell.CreateShortCut(path_link)
        shortcut.Description = f"{pkg_name.upper()} v{__version__}"
        shortcut.TargetPath = target
        shortcut.Arguments = arguments
        shortcut.WorkingDirectory = os.getenv("USERPROFILE")
        shortcut.IconLocation = ico_path
        shortcut.Save()


def unlink_app():
    """
    Remove start menu item and desktop shortcut to `qats` desktop application.
    """
    if not sys.platform == "win32":
        print(f"Unable to remove links to app on {sys.platform} OS.")
        sys.exit()

    from win32com.client import Dispatch

    pkg_name = "qats"
    lnk_name = pkg_name.upper() + ".lnk"

    shell = Dispatch("WScript.Shell")
    for loc in ("Desktop", "Programs"):
        location = shell.SpecialFolders(loc)
        path_link = os.path.join(location, lnk_name)

        if os.path.exists(path_link):
            os.remove(path_link)


def launch_app(home=True, files=None, log_level="info"):
    """
    Start desktop application.

    Parameters
    ----------
    home : bool, optional
        Use home directory as work directory. Else current work directory will be used.
    files : list, optional
        Initialize the application with these time series files (paths).
    log_level: str, optional
        Set logging level, default 'info'
    """
    if home:
        os.chdir(os.getenv("USERPROFILE") or os.getcwd())   # fail safe if no USERPROFILE env

    # install handler for exceptions
    sys.excepthook = handle_exception

    # launch app and main window and deal with possible files
    app = QApplication(sys.argv)
    form = Qats(files_on_init=files, logging_level=log_level)
    form.show()
    sys.exit(app.exec_())


def main():
    """
    Launch desktop application from command line with parameters.
    """
    # top-level parser
    parser = argparse.ArgumentParser(prog="qats",
                                     description="qats is a library and desktop application for time series analysis")
    parser.add_argument("--version", action="version", version=f"qats {__version__}", help="Package version")
    subparsers = parser.add_subparsers(title="Commands", dest="command")

    # app parser
    app_parser = subparsers.add_parser("app", help="Launch the desktop application")
    app_parser.add_argument("-f", "--files", type=str, nargs="*", help="Time series files.")
    app_parser.add_argument("--home", action="store_true",
                            help="Launch from home directory, default: current work directory.")
    app_parser.add_argument("--log-level", default="info", choices=list(LOGGING_LEVELS.keys()),
                            help="Set logging level.")

    # config parser
    config_parser = subparsers.add_parser("config", help="Configure the package")
    applink_group = config_parser.add_mutually_exclusive_group()
    applink_group.add_argument("--link-app", action="store_true",
                               help="Create start menu and destop links to the app (GUI).")
    applink_group.add_argument("--unlink-app", action="store_true",
                               help="Remove start menu and destop links to the app.")

    # parse command line arguments
    args = parser.parse_args()

    if args.command == "app":
        # launch app
        launch_app(home=args.home, files=args.files, log_level=args.log_level)

    elif args.command == "config":
        if args.link_app:
            link_app()
        elif args.unlink_app:
            unlink_app()
        else:
            pass

    else:
        # no arguments given, print the short usage message to assist user
        parser.print_usage()


if __name__ == '__main__':
    main()



"""
Readers for ASCII and direct access formatted time series files exported from SIMA
"""
import fnmatch
import os
import re
from struct import unpack

import numpy as np
from scipy.interpolate import interp1d


def read_sima_wind_names(path):
    """
    Read time series names from key-files associated with '<>_witurb.bin' and '<>_blresp.bin' time-series files
    exported from SIMA/RIFLEF Dynmod.

    Parameters
    ----------
    path: str
        Keyfile path

    Returns
    -------
    list
        Time series names
    """

    # open and read
    with open(path, 'r') as f:
        lines = f.readlines()

    # Extract storage info lines
    p = re.compile(r'ignore*')
    key_lines = [li for li in lines if not li.startswith("'") and not p.search(li.lower())]

    # for each row, split into columns (space-separated) and keep first three
    rows = [lo.split(maxsplit=3)[:3] for lo in key_lines]

    # check which column 'Time' is in
    # note:
    #   idx_time is 1 for the witurb format exported by RIFLEX version <4.20
    #   idx_time is 2 for the witurb format exported by RIFLEX version >=4.20
    idx_time = rows[0].index('Time')

    # pop first row (the one with 'Time')
    _ = rows.pop(0)

    # extract keys according inferred file format
    if idx_time == 1:
        # SIMA/RIFLEX version <4.20: only keys, no wind turbine name
        keys = [row[1] for row in rows]
    else:  # idx_time == 2:
        # SIMA/RIFLEX version >=4.20: wind turbine name in second
        keys = [row[1] + '_' + row[2] for row in rows]

    return keys


def read_names(path):
    """
    Read time series names from key-files associated with .bin and .asc time-series files exported from
    SIMA/RIFLEX Dynmod.

    Parameters
    ----------
    path : str
        Keyfile path

    Returns
    -------
    list
        Time series names

    Warnings
    --------
    This function is under development, and does not presently handle all
    kinds of response listings. If an unknown response description is
    encountered, the key suffix will be `DOF<xx>`.

    """
    # Line id. for pure digit line entries
    linid = 'Lin'
    # zero padding width
    nlin = 2
    nseg = 3
    nnod = 3  # for elements/nodes

    # extract 'noddis', 'elmsfo', 'elmfor' (asc), ...
    keyfiletype = os.path.splitext(path.split('_')[-1])[0].lower()

    # open and read
    with open(path, 'r') as f:
        lines = f.readlines()

    # define parameters
    if keyfiletype == 'noddis':
        elkey = 'No'
    elif keyfiletype in ('elmsfo', 'elmfor'):
        elkey = 'El'
    else:
        elkey = ''
    suffices = _name_suffices(lines)

    # find start line for storage information
    i_start = 1 + lines.index(fnmatch.filter(lines, "*------------------------------------------------------*")[0])
    # extract storage info lines
    p = re.compile(r'ignore*')
    key_lines = [l for l in lines[i_start:] if l.strip() and not p.search(l)]

    # determine number of keys on each line, and in total
    nkeys = [int(line.split()[3]) for line in key_lines]
    nkeystot = sum(nkeys)

    # parse keys
    keys = [''] * nkeystot
    i = 0
    for nk, kl in zip(nkeys, key_lines):
        l = kl.split()
        # Line id.
        if l[0].isdigit():
            a = linid + str(l[0]).zfill(nlin)
        else:
            a = l[0]
        # Segment id.
        b = 'Seg' + str(l[1]).zfill(nseg)
        # El./Node id.
        c = elkey + str(l[2]).zfill(nnod)
        # define key entries
        for suff in suffices[:nk]:
            keys[i] = '_'.join([a, b, c, suff])
            i += 1
    return keys


def read_bin_data(path, ind=None, verbose=False):
    """
    Read time series from binary .bin file exported from RIFLEX Dynmod.

    Parameters
    ----------
    path : str
        Path to .bin file
    ind : list of integers, optional
        Defines which responses to include in returned array. Each of the indices in `ind` refer the sequence number
        of the reponses. Note that `0` is the index of the time array.
    verbose : bool, optional
        Increase verbosity

    Returns
    -------
    array
        Time and data

    Notes
    -----
    If ``ind`` is specified, time array is only included if `0` is included in the specified indices.

    If ``ind`` is specified, the response arrays (1-D) are stored in the returned 2-D array in the same order as
    specified. I.e. if ``ind=[0,10,2,3]``, then the response array with index `10` on .ts file is obtained from
    ``arr[1,:]``.

    """
    if verbose:
        print('Reading %s ...' % path)

    # read
    nbytes = 4
    intfmt = 'i'  # integer format
    with open(path, 'rb') as f:
        # parse no. of records
        nbytesrec = unpack('i', f.read(4))[0]
        nrec = int(nbytesrec / nbytes + 2)  # plus 2 due to first and last cols.
        nts = int(nrec - 2 - 1)             # no. time series (excl. time array)

        # parse no. of time steps
        f.seek(0, 2)                        # go to end
        nbytestot = f.tell()                # get position
        f.seek(0)                           # go back to start
        ndat = int(nbytestot / (nbytes * nrec))

        # which keys/indices to extract
        if ind is None:
            ind = range(nts + 1)            # (including time array)

        # info
        if verbose:
            print('---------------------------------------')
            print('nts (no. of responses on file) : %d' % nts)
            print('ndat (no. of time steps)       : %d' % ndat)
            print('number of keys to read         : %d  (%s)' % (len(ind), "including time array index 0 is specified"))

        # prepare for reading
        basefmt = intfmt + 'f' + 'f' * nts + intfmt
        fmt = basefmt * ndat
        # read, unpack, reshape and transpose (ignoring first and last row
        s = f.read(nbytestot)
        arr = np.array(unpack(fmt, s)).reshape((ndat, nrec))[:, 1:-1].T

    # return specified indices
    return arr[ind, :]


def read_ascii_data(path, ind=None, verbose=False):
    """
    Read time series arranged column wise on ascii formatted file exported from SIMA/RIFLEX Dynmod.

    Parameters
    ----------
    path : str
        Name of the binary file (incl. extension).
    ind : list or tuple, optional
        Which columns to read, with 0 being the first. For example, ind=(1,4,5) will extract the
        2nd, 5th and 6th columns. The default, None, results in all columns being read.
    verbose : bool, optional
        Increase verbosity

    Returns
    -------
    array
        Time and data

    Notes
    -----
    Time is assummed to be in the first column
    If ``ind`` is specified, time array is only included if `0` is included in the specified indices.

    If ``ind`` is specified, the response arrays (1-D) are stored in the returned 2-D array in the same order as
    specified. I.e. if ``ind=[0,10,2,3]``, then the response array with index `10` on ascii file is obtained from
    ``data[1,:]``.

    """
    if verbose:
        print('Reading %s ...' % path)

    # read
    with open(path, 'r') as f:
        # skip commmented lines at start of file
        for line in f:
            if not line.startswith("#"):
                # skip all comment lines
                break

        # load data from the remaining rows as an array
        data = np.loadtxt(f, skiprows=0, usecols=ind, unpack=True)

    # info
    if verbose:
        nts, ndat = data.shape
        print('---------------------------------------')
        print('nts (no. of responses read) : %d' % nts)
        print('ndat (no. of time steps)    : %d' % ndat)

    return data


def write_simo_file(path, time: np.ndarray, data: np.ndarray, dt: float = 0.2, description: str = None):
    """
    Write time series on simo input format

    Parameters
    ----------
    path : str
        File path
    time : array
        Time (s)
    data : array
        Time series data
    dt : float, optional
        Time step on file, default 0.2 seconds
    description : str, optional
        Description printed to file
    """
    # resample to constant dt
    if time[0] < 0:
        duration = time[-1]  # Ignoring everything before 0
    else:
        duration = time[-1] - time[0]
    f = interp1d(time, data)
    trs = np.arange(0., duration, dt)
    drs = f(trs)
    n = np.size(drs)  # reset

    with open(path, "w") as f:
        f.write(f"{n}" + "\n")
        f.write(f"{dt}" + "\n")
        f.write(f"{description}" + "\n" if description is not None else "\n")
        f.write("\n")
        for i in range(n):
            f.write(f"{drs[i]}" + "\n")


def _name_suffices(txt):
    """
    Creates time series name suffices e.g. "Te" and "Mx1", from parsing the "DOF" information in the key-file from
    RIFLEX-DYNMOD.

    Parameters
    ----------
    txt : list
        The lines of the key-file

    Returns
    -------
    list
        Suffices, e.g. 'Xd', 'Yd', or 'Te', 'Sy1', 'Sy2', etc.
    """
    # extract indices for blocks with DOF descriptions
    p = re.compile("following applies")
    inds = [txt.index(line) for line in [line for line in txt if p.search(line)]]

    # Always choose the last block:
    #   If there are two blocks, the first will be bar elements, for which only the axial tension is stored.
    #   This dof will also be the first dof for beam elements, if there are any. Hence, only the last block needs
    #   to be parsed.
    ind = inds[-1]
    p = re.compile("DOF")
    # list of DOF (...) strings
    dofstrings = [line for line in txt[ind:] if p.search(line)]

    # extract DOF descriptions by splitting on '='
    dofdescr = [x.split('=')[-1].strip() for x in dofstrings]

    # define suffices
    suffices = [''] * len(dofdescr)
    for i, ds in enumerate(dofdescr):
        # displacements
        if re.match('displacement', ds):
            # extract direction : x, y, z
            suff = ds.split()[2].upper() + 'd'
        # forces, moments
        elif re.match('Axial', ds):
            suff = 'Te'
        elif re.match('Torsional', ds):
            suff = 'Mx'
        elif re.match('Mom.', ds):
            dsp = ds.split()
            suff = 'M'
            # which axis
            suff += fnmatch.filter(dsp, '*axis*')[0][0]
            # which end
            suff += str(dsp[-1])
        elif re.match('Shear', ds):
            dsp = ds.split()
            suff = 'S'
            # which axis
            suff += fnmatch.filter(dsp, '*direction*')[0][0]
            # which end
            suff += str(dsp[-1])

        else:  # unknown description, use DOFxx
            suff = 'DOF' + str(i + 1).zfill(2)

        suffices[i] = suff

    return suffices





import sys
import traceback

from qtpy.QtCore import QObject, QRunnable
from qtpy.QtCore import Signal as QSignal
from qtpy.QtCore import Slot as QSlot


class WorkerSignals(QObject):
    """
    Defines the signals available from a running worker thread.

    Supported signals are:

    finished
        No data

    error
        `tuple` (exctype, value, traceback.format_exc() )

    result
        `object` data returned from processing, anything
    """
    finished = QSignal()
    error = QSignal(tuple)
    result = QSignal(object)


class Worker(QRunnable):
    """
    Worker thread inherits from QRunnable to handler worker thread setup, signals and wrap-up.

    Parameters
    ----------
    callback: function
        The function callback to run on this worker thread. Supplied args and kwargs will be passed through to the runner.
    *args
        Variable length argument list passed to callback function.
    **kwargs
        Arbitrary keyword arguments passed to callback function.
    """
    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()
        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

    @QSlot()
    def run(self):
        """
        Initialise the runner function with passed args, kwargs.
        """
        # Retrieve args/kwargs here; and fire processing using them
        try:
            result = self.fn(*self.args, **self.kwargs)
        except:
            # catch exception and return traceback
            exctype, value = sys.exc_info()[:2]
            self.signals.error.emit((exctype, value, traceback.format_exc()))
        else:
            # Return the result of the processing
            self.signals.result.emit(result)
        finally:
            # Done
            self.signals.finished.emit()



# -*- coding: utf-8 -*-
"""
Sub-package with io for various file formats.
"""
from . import csv, direct_access, other, sima, sima_h5
from . import sintef_mat  # for backwards compatibility
from . import sintef_mat as matlab



import os
from typing import List, Tuple, Union

from nptdms import TdmsFile


def read_names(path):
    """
    Read time series names from Technical Data Management Streaming (TDMS) file.

    Parameters
    ----------
    path: str
        Path (relative or absolute) to tdms-file

    Returns
    -------
    list
        List of time series names (datasets)

    References
    ----------
    1. http://www.ni.com/product-documentation/3727/en/

    """
    if not os.path.isfile(path):
        raise FileNotFoundError("file not found: %s" % path)

    f = TdmsFile(path)
    names = [f"{g.name}\\{c.name}" for g in f.groups() for c in f[g.name].channels() if c.name.lower() != 'time']

    return names


def read_data(path: str, names: Union[List[str], Tuple[str]] = None):
    """
    Read time series data from Technical Data Management Streaming (TDMS) files typically exported from LabVIEW.

    Parameters
    ----------
    path: str
        File name.
    names: str or list, optional
        Timeseries/dataset names. If None (default), all time series are read.

    Returns
    -------
    list
        List of arrays with time and data

    """
    if not os.path.isfile(path):
        raise FileNotFoundError("file not found: %s" % path)

    if isinstance(names, str):
        names = [names]
    elif type(names) in (list, tuple):
        pass
    elif names is None:
        # if names not specified, get all names
        names = read_names(path)
    else:
        raise TypeError("`names` must be str/list/tuple, got: %s" % type(names))

    arrays = []
    with TdmsFile.open(path) as f:
        for name in names:
            # assuming an object hierarchy with depth 2 'group-channel'
            assert len(name.split("\\")) == 2, f"Unable to parse group name and channel name from {name}."
            group_name, channel_name = name.split("\\")
            group = f[group_name]
            channel = group[channel_name]
            data = channel[:]
            time = None

            try:
                # try to fetch time track defined by wf_start_time and wf_start_offset attributes
                time = channel.time_track()
            except KeyError:
                # wf_start_time and wf_start_offset attributes does not exist
                # check if time is a separate channel
                for _ in ["time", "Time"]:
                    try:
                        time = group[_][:]
                    except KeyError:
                        pass
                    else:
                        break
            finally:
                if time is None:
                    raise KeyError(f"Could not find time array for channel {channel_name}.")
                elif not time.shape == data.shape:
                    # verify that time and data arrays have the same shape
                    raise ValueError(f"Time {time.shape} and data {data.shape} are not equally shaped.")
                else:
                    arrays.append([time, data])

    return arrays



#!/usr/bin/env python
# encoding: utf8
"""
Module with functions for signal processing.
"""
import warnings

import numpy as np
from scipy.fftpack import fft, ifft, irfft, rfft
from scipy.signal import butter
from scipy.signal import coherence as spcoherence
from scipy.signal import csd as spcsd
from scipy.signal import filtfilt, sosfiltfilt, welch


def extend_signal_ends(x: np.ndarray, n: int) -> np.ndarray:
    """Extend the signal ends with `n` values to mitigate the edge effect.

    Parameters
    ----------
    x : array_like
        Signal
    n : int
        Number of values prepended and appended to signal.

    Notes
    -----
    At each end of the signal `n` values of the signal are replicated, flipped and joined with the signal to maintain
    continuity in the signal level and slope at the joining points. This should mitigate end effects when filterin
    the signal.

    The original signal is retrieved as `x[n:-n:1]`.
    """
    start = 2. * x[0] - 1. * x[n:0:-1]
    end = 2. * x[-1] - 1. * x[-2:-(n + 2):-1]
    return np.concatenate((start, x, end))


def smooth(x: np.ndarray, window_len: int = 11, window: str = 'rectangular', mode: str = 'same') -> np.ndarray:
    """
    Smooth time serie based on convolution of a window function and the time serie.

    Parameters
    ----------
    x : array
        The input signal.
    window_len : int, optional
        The dimension of the smoothing window.
    window : {'rectangular', 'hanning', 'hamming', 'bartlett', 'blackman'}, optional
        The type of window. Rectangular window will produce a moving average smoothing.
    mode : {‘same’, ‘valid’, ‘full’}, optional
        full:
           This returns the convolution at each point of overlap, with an output
           shape of (N+M-1,). At the end-points of the convolution, the signals
           do not overlap completely, and boundary effects may be seen.

        same:
           By default mode is 'same' which returns output of length max(M, N).
           Boundary effects are still visible.

        valid:
           Mode valid returns output of length max(M, N) - min(M, N) + 1. The
           convolution product is only given for points where the signals overlap
           completely. Values outside the signal boundary have no effect

    Returns
    -------
    array
        The smoothed signal.

    Notes
    -----
    This method is based on the convolution of a scaled window with the signal. The signal is prepared by introducing
    reflected copies of the signal (with the window size) in both ends so that transient parts are minimized
    in the beginning and end of the output signal.

    Examples
    --------
    >>> from numpy import linspace
    >>> from numpy.random import randn
    >>> t = linspace(-2,2,0.1)
    >>> x = sin(t)+randn(len(t))*0.1
    >>> y = smooth(x)

    References
    ----------
    1. Wikipedia, http://en.wikipedia.org/wiki/Convolution

    See Also
    --------
    numpy.convolve

    """
    x = np.asarray(x)

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if window not in ('rectangular', 'hanning', 'hamming', 'bartlett', 'blackman'):
        raise ValueError("Window is not one of '{0}', '{1}', '{2}', '{3}', '{4}'".format(
                   *('rectangular', 'hanning', 'hamming', 'bartlett', 'blackman')))

    if window == 'rectangular':  # moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.' + window + '(window_len)')

    if mode == 'valid':
        s = np.r_[x[window_len - 1:0:-1], x, x[-1:-window_len:-1]]
        y = np.convolve(w / w.sum(), s, mode=mode)
    else:
        s = np.r_[2 * x[0] - x[window_len:1:-1], x, 2 * x[-1] - x[-1:-window_len:-1]]
        y = np.convolve(w / w.sum(), s, mode=mode)
        y = y[window_len - 1:-window_len + 1]

    return y


def taper(x: np.ndarray, window: str = 'tukey', alpha: float = 0.001) -> (np.ndarray, float):
    """
    Taper the input time serie using a window function

    Parameters
    ----------
    x : array
        Time series (without time vector), dimension `n*1`.
    window : {'tukey','cosine','hanning', 'flat', ....}
       Window function type. See numpy documentation for more windows
    alpha : float, optional
        Fraction of time domain signal to be tapered. Applies only
        to tukey and kaiser windows.

    Returns
    -------
    array
        Tapered time domain signal
    float
        correction factor to prevent FFT components from diminishing after the windowing.

    Notes
    -----
    All FFT based measurements assume that the signal is periodic in the time frame. When the measured signal is
    not periodic then leakage occurs. Leakage results in misleading information about the spectral amplitude and
    frequency. A window is shaped so that it is exactly zero at the beginning and end of the data block and has
    some special shape in between. This function is then multiplied with the time data block forcing the signal to be
    periodic and ultimately reduces the effects of leakage. There are many windows to choose from, each with advantages
    for specific applications. You must understand the effects of leakage and know the tradeoffs and advantages of the
    various windowing functions to accurately interpret frequency domain measurements.

    The cosine window is also known as the sine window.

    The Tukey window is also known as the tapered cosine window.

    See Also
    --------
    numpy.bartlett, numpy.blackman, numpy.hamming, numpy.hanning, numpy.kaiser

    References
    ----------
    1. Wikipedia, http://en.wikipedia.org/wiki/Window_function
    2. Melbourne G. Briscoe (1972), Energy loss in surface wave spectra due to data windowing, North Atlantic Treaty Organization (NATO), Saclant ASW Research Centre,

    """
    window_len = np.size(x)
    window = window.lower()

    # choice of window function
    if window == 'rectangular':
        w = np.ones(window_len)
    elif window == 'tukey':
        # alpha = 0 - rectangular window, alpha - Hann window
        w = np.zeros(window_len)
        for i in range(window_len):
            if (i >= 0) & (i < alpha * window_len / 2):
                w[i] = 0.5 * (1 + np.cos(np.pi * (2 * i / (alpha * window_len) - 1)))
            if (i >= alpha * window_len / 2) & (i <= window_len * (1 - alpha / 2)):
                w[i] = 1
            if (i > window_len * (1 - alpha / 2)) & (i <= window_len):
                w[i] = 0.5 * (1 + np.cos(np.pi * (2 * i / (alpha * window_len) - 2 / alpha + 1)))
    elif window == 'cosine':
        # also known as sine window
        n = np.arange(window_len)
        w = np.sin(np.pi * n / (window_len - 1))
    elif window == 'kaiser':
        w = eval('np.' + window + '(window_len,alpha)')
    else:
        w = eval('np.' + window + '(window_len)')

    # calculate tapered time series
    y = x * w

    # calculate weighting factor that should be applied so that the correct FFT signal amplitude level is recovered
    # after the windowing.
    wcorr = np.sum(w ** 2) / window_len

    return y, wcorr


def lowpass(x: np.ndarray, dt: float, fc: float, order: int = 5) -> np.ndarray:
    """
    Low pass filter data signal x at cut off frequency fc, blocking harmonic content above fc.
    
    Parameters
    ----------
    x : array_like
        Signal
    dt : float
        Signal sampling rate (s)
    fc : float
        Cut off frequency (Hz)
    order : int, optional
        Butterworth filter order. Default 5.
    
    Returns
    -------
    array
        Filtered signal

    See Also
    --------
    scipy.signal.butter, scipy.signal.filtfilt
    """
    nyq = 0.5 * 1. / dt         # nyquist frequency
    normal_cutoff = fc / nyq    # normalized cut off frequency
    b, a = butter(order, normal_cutoff, btype='lowpass', analog=False)
    y = filtfilt(b, a, x)
    return y


def highpass(x: np.ndarray, dt: float, fc: float, order: int = 5) -> np.ndarray:
    """
    High pass filter data signal x at cut off frequency fc, blocking harmonic content below fc.
    
    Parameters
    ----------
    x : array_like
        Signal
    dt : float
        Signal sampling rate (s)
    fc : float
        Cut off frequency (Hz)
    order : int, optional
        Butterworth filter order. Default 5.
    
    Returns
    -------
    array
        Filtered signal

    See Also
    --------
    scipy.signal.butter, scipy.signal.filtfilt
    """
    nyq = 0.5 * 1. / dt         # nyquist frequency
    normal_cutoff = fc / nyq    # normalized cut off frequency
    b, a = butter(order, normal_cutoff, btype='highpass', analog=False)
    y = filtfilt(b, a, x)
    return y


def bandpass(x: np.ndarray, dt: float, flow: float, fupp: float, order: int = 5) -> np.ndarray:
    """
    Band pass filter data signal x at cut off frequencies flow and fupp, blocking harmonic content outside the
    frequency band [flow, fupp]
    
    Parameters
    ----------
    x : array_like
        Signal
    dt : float
        Signal sampling rate (s)
    flow, fupp : float
        Passing frequency band (Hz)
    order : int, optional
        Butterworth filter order. Default 5.
    
    Returns
    -------
    array
        Filtered signal

    See Also
    --------
    scipy.signal.butter, scipy.signal.sosfiltfilt
    """
    nyq = 0.5 * 1. / dt  # nyquist frequency
    normal_cutoff = (flow / nyq, fupp / nyq)  # normalized cut off frequencies
    sos = butter(order, normal_cutoff, btype='bandpass', analog=False, output='sos')
    y = sosfiltfilt(sos, x)
    return y


def bandblock(x: np.ndarray, dt: float, flow: float, fupp: float, order: int = 5) -> np.ndarray:
    """
    Band block filter data signal x at cut off frequencies flow and fupp, blocking harmonic content inside the
    frequency band [flow, fupp]
    
    Parameters
    ----------
    x : array_like
        Signal
    dt : float
        Signal sampling rate (s)
    flow, fupp : float
        Blocked frequency band (Hz)
    order : int, optional
        Butterworth filter order. Default 5.
    
    Returns
    -------
    array
        Filtered signal

    Notes
    -----
    SciPy bandpass/bandstop filters designed with b, a are unstable and may result in erroneous filters at higher
    filter orders. Here we use sos (second-order sections) output of filter design instead.

    See Also
    --------
    scipy.signal.butter, scipy.signal.sosfiltfilt
    """
    nyq = 0.5 * 1. / dt                         # nyquist frequency
    normal_cutoff = (flow / nyq, fupp / nyq)    # normalized cut off frequencies
    sos = butter(order, normal_cutoff, btype='bandstop', analog=False, output='sos')
    y = sosfiltfilt(sos, x)
    return y


def threshold(x: np.ndarray, thresholds: tuple) -> np.ndarray:
    """
    Allow only frequency components whose amplitudes are between the lower threshold value and the upper threshold
    value to pass.
    
    Parameters
    ----------
    x : array_like
        input data signal
    thresholds : tuple
        passing amplitude range, thresholds as fraction of maximum frequency component amplitude
    
    Returns
    -------
    array
        filtered data signal

    Notes
    -----
    FFT filter.

    See Also
    --------
    scipy.fftpack
    """
    real_signal = np.all(np.isreal(x))
    n = x.size
    nfft = int(pow(2, np.ceil(np.log(n) / np.log(2))))
    lth, uth = thresholds   # unpack lower and upper thresholds of passing range

    if real_signal:
        fa = rfft(x, nfft)
        h = np.zeros(np.shape(fa))
        h[(lth*max(abs(fa)) < abs(fa)) & (abs(fa) <= uth*max(abs(fa)))] = 1.0
        x1 = irfft(fa * h, nfft)
    else:
        fa = fft(x, nfft)
        h = np.zeros(np.shape(fa))
        h[(lth*max(abs(fa)) < abs(fa)) & (abs(fa) <= uth*max(abs(fa)))] = 1.0
        x1 = ifft(fa * h, nfft)

    return x1[:n]


def autocorrelation(series: np.ndarray) -> list:
    """
    Estimation of the auto-correlation coefficients of *series*

    Parameters
    ----------
    series : array_like
        data series

    Returns
    -------
    list
        arrays of autocorrelation coefficients for the entire *series* for lags in the range [dt, dt, duration]

    Notes
    -----
    I took a part of code from pandas autocorrelation_plot() function. I checked the answers and the values are
    matching exactly.
    The auto-correlation coefficients can be plotted against the time vector associated with series.

    References
    ----------
    1. Wikipedia, http://en.wikipedia.org/wiki/Autocorrelation

    """
    n = len(series)
    data = np.asarray(series)
    mean = np.mean(data)
    c0 = np.sum((data - mean) ** 2) / float(n)

    def r(h):
        """
        Calculation of autocorrelation coefficients for lag *h*

        Parameters
        ----------
        h : float
            lag

        Returns
        -------
        array
           autocorrelation coefficients for lag *h*

        """
        acf_lag = ((data[:n - h] - mean) * (data[h:] - mean)).sum() / float(n) / c0
        return round(acf_lag, 3)

    x = np.arange(n)  # Avoiding lag 0 calculation
    return list(map(r, x))


def average_frequency(t: np.ndarray, x: np.ndarray, up: bool = True) -> float:
    """
    Average frequency of mean level crossings.

    Parameters
    ----------
    t : array_like
        Time (seconds).
    x : array_like
        Signal.
    up : bool, optional
        - True: Period based on average time between up-crossings
        - False: Period based on average time between down-crossings


    Returns
    -------
    float
        Average frequency of mean level crossings (Hz)

    """
    # remove mean value from time series
    x_ = x - np.mean(x)

    if up:
        crossings = 1 * (x_ > 0.)
        indicator = 1
    else:
        crossings = 1 * (x_ < 0.)
        indicator = -1

    crossings = np.diff(crossings)  # array with value=1 at position of each up-crossing and -1 at each down-crossing
    crossings[crossings != indicator] = 0   # remove crossings with opposite direction
    i = np.where(crossings == indicator)[0] + 1  # indices for crossings
    d = (t[i[-1]] - t[i[0]]) / (np.abs(np.sum(crossings)) - 1)  # duration between first and last crossing
    return 1./d


def find_maxima(x, local: bool = False, threshold: float = None, up: bool = True) -> (np.ndarray, np.ndarray):
    """
    Return sorted maxima

    Parameters
    ----------
    x : array
        Signal.
    local : bool, optional
        If True, local maxima are also included (see notes below). Default is to include only global maxima.
    threshold : float, optional
        Include only maxima larger than specified treshold. Default is not to remove any of the maxima identified.
    up : bool, optional
        This parameter is deprecated and has no effect on the peak/maxima identification.

    Returns
    -------
    array
        Signal maxima, sorted from smallest to largest.
    array
        Indices of signal maxima.

    Notes
    -----
    By default only 'global' maxima are considered, i.e. the largest maximum between each mean-level up-crossing.
    If ``local=True``, local maxima are also included (first derivative is zero, second derivative is negative).

    *Changes from version 4.12.*
    When extracting global maxima (``local=False``, the default), positive half-cycles at the beginning or end of the
    time series are included. For example, if there is a mean-level up-crossing before the first down-crossing, the
    maximum value between these crossings is included. Similarly, if there is a down-crossing after the last
    up-crossing, the maximum value between these crossings is included. This also implies that there is no difference
    in considering mean level up-crossings (`up=True`) vs. down-crossings (`up=False`). The `up` parameter is
    therefore deprecated.

    Examples
    --------
    Extract global maxima from time series signal `x`:

    >>> maxima, _ = find_maxima(x)

    Extract global maxima and corresponding indices:

    >>> maxima, indices = find_maxima(x)

    Assuming `time` is the time vector (numpy array) for signal `x`, the following example will provide an array of
    time instants associated with the maxima sample:

    >>> time_maxima = time[indices]

    Note that the returned maxima (and corresponding indices) are sorted in ascending order. To reverse this and obtain
    an array with the maxima arranged in order of appearance, the following code may be used:

    >>> indsort = np.argsort(indices)
    >>> maxima = maxima[indsort]
    >>> indices = indices[indsort]
    >>> time_maxima = time[indices]

    """
    # parameter `up` is deprecated and has no effect
    if up is not True:
        warnings.warn("qats.signal.find_maxima: parameter `up` is deprecated and has no effect", category=DeprecationWarning)
    
    # remove mean value from time series to identify crossings
    x_ = x - np.mean(x)

    # find maxima
    if not local:
        # global maxima (largest between mean-level crossings)
        # identify crossings
        crossings = 1 * (x_ > 0.)
        crossings = np.diff(crossings)  # array with 1 at position of each up-crossing and -1 at each down-crossing

        # get array indices for up-/down-crossings
        crossing_indices_up = np.where(crossings == 1)[0] + 1   # up-crossings
        crossing_indices_do = np.where(crossings == -1)[0] + 1  # down-crossings

        # use up-crossings as the basis
        crossing_indices = crossing_indices_up

        # if there is a down-crossing after the last up-crossing, add that crossing as well
        # (this is to avoid losing the last peak, and is particularly important for problems with few crossings,
        #  e.g., due to low-frequent oscillations)
        if crossing_indices_up[-1] < crossing_indices_do[-1]:
            crossing_indices = np.append(crossing_indices, crossing_indices_do[-1])

        # number of crossings and number of peaks
        n_crossings = crossing_indices.size
        n_peaks = n_crossings - 1

        # no global maxima if the signal crosses mean only once
        if n_crossings < 2:
            return np.array([]), np.array([], dtype=int)

        # initiate arrays to be populated subsequently
        maxima = np.zeros(n_peaks)
        maxima_indices = np.zeros(n_peaks, dtype=int)

        # loop to find max. between each up-crossing:
        for j, start in enumerate(crossing_indices[:-1]):
            stop = crossing_indices[j + 1]
            maxima[j] = x[start:stop].max()
            maxima_indices[j] = start + np.argmax(x[start:stop])

    else:
        # local maxima (all peaks)
        ds = 1 * (np.diff(x) < 0)     # zero while ascending (positive derivative) and 1 while descending
        ds = np.append(ds, [0])       # lost data points when differentiating, close cycles by adding 0 at end
        d2s = np.diff(ds)             # equal to +/-1 at each turning point, +1 indicates maxima
        d2s = np.insert(d2s, 0, [0])  # lost data points when differentiating, close cycles by adding 0 at start

        maxima_indices = np.where(d2s == 1)[0]  # unpack tuple returned from np.where
        maxima = x[maxima_indices]

        n_peaks = maxima.size

        # return quickly if no peaks/local maxima were found (e.g., if time series is monotonically increasing)
        if n_peaks == 0:
            return np.array([]), np.array([], dtype=int)

    # discard maxima lower than specified threshold
    if threshold is not None:
        above_threshold = (maxima >= threshold)
        maxima = maxima[above_threshold]
        maxima_indices = maxima_indices[above_threshold]
    else:
        pass

    # sort ascending
    ascending = np.argsort(maxima)
    maxima = maxima[ascending]
    maxima_indices = maxima_indices[ascending]

    return maxima, maxima_indices


def psd(x: np.ndarray, dt: float, **kwargs) -> (np.ndarray, np.ndarray):
    """
    Estimate power spectral density of discrete time signal X using Welch’s method.

    Parameters
    ----------
    x : array_like
        Time series data.
    dt : float
        Time step.
    kwargs : optional
        See `scipy.signal.welch` documentation for available options.

    Returns
    -------
    array
        Frequencies
    array
        Corresponding power spectral density

    Notes
    -----
    This function basically wraps `scipy.signal.welch` to control defaults etc.

    See also
    --------
    scipy.signal.welch, scipy.signal.periodogram

    """
    x = np.asarray(x)

    # estimate psd using welch's definition
    f, p = welch(x, fs=1./dt, **kwargs)

    return f, p


def csd(x: np.ndarray, y: np.ndarray, dt: float, **kwargs) -> (np.ndarray, np.ndarray):
    """
    Estimate cross power spectral density of discrete-time signals X and Y using Welch’s method.

    Parameters
    ----------
    x : array_like
        Time series data.
    y : array_like
        Time series data.
    dt : float
        Time step.
    kwargs : optional
        See `scipy.signal.csd` documentation for available options.

    Returns
    -------
    array
        Frequencies
    array
        Corresponding cross power spectral density

    Notes
    -----
    This function basically wraps `scipy.signal.csd` to control defaults etc.

    See also
    --------
    scipy.signal.welch, scipy.signal.csd

    """
    x = np.asarray(x)
    y = np.asarray(y)

    # estimate csd using welch's definition
    f, p = spcsd(x, y, fs=1./dt, **kwargs)

    return f, p


def coherence(x: np.ndarray, y: np.ndarray, dt: float, **kwargs) -> (np.ndarray, np.ndarray):
    """
    Estimate the magnitude squared coherence estimate of discrete-time signals X and Y using Welch’s method.

    Parameters
    ----------
    x : array_like
        Time series data.
    y : array_like
        Time series data.
    dt : float
        Time step.
    kwargs : optional
        See `scipy.signal.coherence` documentation for available options.

    Returns
    -------
    array
        Frequencies
    array
        Corresponding cross power spectral density

    Notes
    -----
    This function basically wraps `scipy.signal.coherence` to control defaults etc.

    See also
    --------
    scipy.signal.welch, scipy.signal.coherence

    """
    x = np.asarray(x)
    y = np.asarray(y)

    # estimate coherence using welch's definition
    f, c = spcoherence(x, y, fs=1./dt, **kwargs)

    return f, c


def tfe(x: np.ndarray, y: np.ndarray, dt: float, clim: float = None, **kwargs) -> (np.ndarray, np.ndarray):
    """
    Estimate the transfer function between two discrete-time signals X and Y using Welch’s method.

    Parameters
    ----------
    x : array_like
        Time series data.
    y : array_like
        Time series data.
    dt : float
        Time step.
    clim : float, optional
        Discard transfer function estimates where the magnitude squared coherence estimate is below this limit.
    kwargs : optional
        See `scipy.signal.welch`, `scipy.signal.csd` and `scipy.signal.coherence` documentation for available options.

    Returns
    -------
    array
        Frequencies
    array
        Corresponding transfer function estimate

    Examples
    --------
    Estimate the transfer function between wave elevation and vessel heave motion recorded to a file "somefile.mat".
    >>> from qats import TsDB
    >>> from qats.signal import tfe
    >>> import matplotlib.pyplot as plt
    >>>
    >>> db = TsDB.fromfile("somefile.mat")
    >>> wave = db.get("wave_2")
    >>> heave = db.get("heave")
    >>>
    >>> # discard the transient signal
    >>> t, w = wave.get(twin=(1000, 1e8))
    >>> _, h = heave.get(twin=(1000, 1e8))
    >>> dt = t[1] - t[0]
    >>> # discard part of signals with poor coherence and smooth using Welch's method with 1000 values per segment.
    >>> f, tf = tfe(w, h, dt, clim=0.3, nperseg=1000)
    >>> # plot transfer function against period and limit to periods larger than 2 seconds (0.5Hz)
    >>> plt.plot(1. / f[(0. < f) & (f <= 0.5)], abs(tf[(0. < f) & (f <= 0.5)]))
    >>> plt.xlabel("Period (Hz)")
    >>> plt.ylabel("Transfer function (-)")
    >>> plt.grid()
    >>> plt.show()

    Notes
    -----
    For single input/single-output systems like this the transfer function is estimated as Pyx / Pxx where Pxx is the
    power spectral density of x and Pyx is the complex conjugate of the cross power spectral density of x and y.

    See also
    --------
    scipy.signal.welch, scipy.signal.coherence

    """
    x = np.asarray(x)
    y = np.asarray(y)

    f, pxx = psd(x, dt, **kwargs)
    _, pyx = csd(x, y, dt, **kwargs)
    tf = pyx / pxx

    if clim is not None:
        _, cyx = coherence(x, y, dt, **kwargs)
        return f[cyx >= clim], tf[cyx >= clim]
    else:
        return f, tf
